"""SDC location handling."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING
from urllib.parse import ParseResult, parse_qsl, quote, unquote, urlencode, urlsplit, urlunparse

if TYPE_CHECKING:
    from collections.abc import Iterable

    from sdc11073.wsdiscovery.service import Service


class UrlSchemeError(Exception):
    """Indicate that the scheme is wrong."""


class SdcLocation:
    """SdcLocation contains all parameters that define a location.

    It serves two purposes:
    1. it can create a scope string for discovery.
    2. it can filter services that are located "inside" this location (this is different from discovery matching!).
        E.g. if this location has only facility and building set, then all services with the same facility and
        building are considered "inside".
        Hierarchy is 'fac', 'bldng', 'flr', 'poc', 'rm', 'bed'.
        It is debatable if a floor contains 1...n points of care or a point of care contains 1...n floors,
        so be careful with this!
    """

    scheme = 'sdc.ctxt.loc'  # constant for scheme from sdc standard

    url_elements = ('fac', 'bldng', 'flr', 'poc', 'rm', 'bed')  # order also defines hierarchy

    def __init__(  # noqa: PLR0913
        self,
        fac: str | None = None,
        poc: str | None = None,
        bed: str | None = None,
        bldng: str | None = None,
        flr: str | None = None,
        rm: str | None = None,
        root: str = 'sdc.ctxt.loc.detail',
    ):
        self._root = root
        self.fac = fac  # facility
        self.bldng = bldng  # building
        self.poc = poc  # point of care
        self.flr = flr  # floor
        self.rm = rm  # room
        self.bed = bed  # Bed

    @property
    def root(self) -> str:
        """Return the root of the location, e.g. 'sdc.ctxt.loc.detail'."""
        warnings.warn('Will be removed without replacement in future version', DeprecationWarning, stacklevel=2)
        return self._root

    @root.setter
    def root(self, value: str):
        warnings.warn('Will be removed without replacement in future version', DeprecationWarning, stacklevel=2)
        self._root = value

    @property
    def scope_string(self) -> str:
        """Return a string that can be used in ScopesType for discovery.

        Example: sdc.ctxt.loc:/sdc.ctxt.loc.detail/HOSP1%2F%2F%2FCU1%2F%2FBed42?fac=HOSP1&poc=CU1&bed=Bed42
        """
        warnings.warn('Will be removed without replacement in future version', DeprecationWarning, stacklevel=2)
        identifiers = []
        query_dict = {}
        for url_name in self.url_elements:
            value = getattr(self, url_name) or ''  # None value becomes an empty string.
            identifiers.append(value)
            if value:
                query_dict[url_name] = value
        identifiers = [quote(ident, safe='') for ident in identifiers]
        slash = quote('/', safe='')
        loc = slash.join(identifiers)  # this is a bit ugly, but urllib.quote does not touch slashes;
        query = urlencode(query_dict)
        path = f'/{quote(self.root)}/{loc}'
        return urlunparse(
            ParseResult(scheme=self.scheme, netloc=None, path=path, params=None, query=query, fragment=None),
        )

    def filter_services_inside(self, services: Iterable[Service]) -> list[Service]:
        """Return services that are 'inside' own location (see doc string of class)."""
        return [s for s in services if self._service_matches(s)]

    def _service_matches(self, service: Service) -> bool:
        if service.scopes is None:
            return False
        return any(self._scope_string_matches(scope) for scope in service.scopes.text)

    def _scope_string_matches(self, scope_text: str) -> bool:
        """Check if location in scope is inside own location."""
        try:
            other = self.__class__.from_scope_string(scope_text)
        except UrlSchemeError:
            # Scope has different scheme, no match
            return False
        else:
            return other in self

    def __contains__(self, other: SdcLocation) -> bool:
        """Compare element by element, root included.

        If own element is None, every value of other is accepted, else they must be identical.
        """
        if self.root != other.root:
            return False
        for attr_name in self.url_elements:
            my_attr = getattr(self, attr_name)
            if my_attr is not None:
                if my_attr != getattr(other, attr_name):
                    return False
        return True

    @classmethod
    def from_scope_string(cls, scope_string: str) -> SdcLocation:
        """Construct a Location from a scope string.

        If url scheme is not 'sdc.ctxt.loc', an UrlSchemeError is raised
        :param scope_string: an url
        :return: a SdcLocation object.
        """
        src = urlsplit(scope_string)

        if src.scheme.lower() != cls.scheme:
            msg = f'scheme "{src.scheme}" not excepted, must be "{cls.scheme}"'
            raise UrlSchemeError(msg)
        dummy, root, _ = src.path.split('/')
        root = unquote(root)
        query_dict = dict(parse_qsl(src.query))
        # make a new argumentsDict with well known keys.
        # This allows to ignore unknown keys that might be present in query_dict
        arguments_dict = {}
        for attr_name in cls.url_elements:
            arguments_dict[attr_name] = query_dict.get(attr_name)

        arguments_dict['root'] = root
        return cls(**arguments_dict)

    def __eq__(self, other: object) -> bool:
        attr_names = (*self.url_elements, 'root')
        try:
            return all(getattr(self, attr_name) == getattr(other, attr_name) for attr_name in attr_names)
        except AttributeError:
            return False

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __str__(self) -> str:
        return f'{self.__class__.__name__} {self.scope_string}'

    def __hash__(self):
        attr_names = (*self.url_elements, 'root')
        return hash(tuple([getattr(self, attr_name) for attr_name in attr_names]))



"""The module defines the interface of a provider mdib."""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from contextlib import AbstractContextManager

    from sdc11073.definitions_base import AbstractDataModel, BaseDefinitions

    from .entityprotocol import ProviderEntityGetterProtocol
    from .transactionsprotocol import (
        ContextStateTransactionManagerProtocol,
        DescriptorTransactionManagerProtocol,
        StateTransactionManagerProtocol,
    )


class ProviderMdibProtocol(Protocol): # pragma: no cover
    """The interface of a provider mdib.

    This interface only expects the ProviderEntityGetterProtocol.
    The old implementation with separate lookups for descriptors, states and context states
    is not part of this protocol.
    """

    entities: ProviderEntityGetterProtocol
    sdc_definitions: type[BaseDefinitions]
    data_model: AbstractDataModel

    def descriptor_transaction(self) -> AbstractContextManager[DescriptorTransactionManagerProtocol]:
        """Return a transaction."""

    def context_state_transaction(self) -> AbstractContextManager[ContextStateTransactionManagerProtocol]:
        """Return a transaction."""

    def alert_state_transaction(self, set_determination_time: bool = True) \
            -> AbstractContextManager[StateTransactionManagerProtocol]:
        """Return a transaction."""

    def metric_state_transaction(self, set_determination_time: bool = True) \
            -> AbstractContextManager[StateTransactionManagerProtocol]:
        """Return a transaction."""

    def rt_sample_state_transaction(self, set_determination_time: bool = False) \
            -> AbstractContextManager[StateTransactionManagerProtocol]:
        """Return a transaction."""

    def component_state_transaction(self) -> AbstractContextManager[StateTransactionManagerProtocol]:
        """Return a transaction."""

    def operational_state_transaction(self) -> AbstractContextManager[StateTransactionManagerProtocol]:
        """Return a transaction."""



from __future__ import annotations

import contextlib
from collections import defaultdict
from typing import TYPE_CHECKING, List, Optional, Union
from .porttypebase import DPWSPortTypeBase
from .porttypebase import WSDLMessageDescription, WSDLOperationBinding
from .porttypebase import mk_wsdl_two_way_operation, msg_prefix
from sdc11073.dispatch import DispatchKey
from sdc11073.namespaces import PrefixesEnum

if TYPE_CHECKING:
    from sdc11073.xml_types.pm_types import LocalizedText

def _tw2i(text_width_string):
    """ text width to int"""
    lookup = {'xs': 0, 's': 1, 'm': 2, 'l': 3, 'xl': 4, 'xxl': 5, None: 999}
    return lookup[text_width_string]


def _calc_number_of_lines(text):
    # definition of a line in Participant Model:
    # ...a line is defined as the content of the text from either the beginning of the text or the beginning of
    # a previous line until the next occurrence of period mark, question mark, exclamation mark, or paragraph.
    # TBD: naive approach?
    return len(text.split('\n'))


def _text_width_filter(localized_texts, width: int):
    candidates = [v for v in localized_texts if _tw2i(v.TextWidth) <= width]
    if candidates:
        candidates.sort(key=lambda obj: _tw2i(obj.TextWidth) or -1)
    return candidates


def _n_o_l_filter(localized_texts, n_o_l):
    candidates = [v for v in localized_texts if v.n_o_l <= n_o_l]
    if candidates:
        candidates.sort(key=lambda obj: obj.n_o_l or -1)
    return candidates


class LocalizationStorage:
    def __init__(self, localized_texts: Optional[List[LocalizedText]] = None):
        self._localized_texts = defaultdict(list)  # key = handle, value = list of LocalizedText objects
        if localized_texts:
            self.add(*localized_texts)

    def add(self, *localized_texts: LocalizedText):
        for text in localized_texts:
            self._localized_texts[text.Ref].append(text)

    def filter_localized_texts(self, requested_handles: Union[List[str], None],
                               requested_version: Union[int, None],
                               requested_langs: Union[List[str], None],
                               text_widths: Union[List[str], None],
                               number_of_lines: Union[List[int], None]):
        """

        :param requested_handles: list of handles
        :param requested_version: an integer or None
        :param requested_langs: list of language strings
        :param text_widths: a list of chars (s, xs, ...)
        :param number_of_lines: a list of integers, 0...n
        :return: a list of LocalizedText instances
        """
        # make integers for text_widths and number_of_lines
        if text_widths is None:
            text_widths = []

        i_text_widths = [_tw2i(w) for w in text_widths]

        if number_of_lines is None:
            number_of_lines = []
        if requested_handles is None:
            requested_handles = []
        i_nls = [int(line) for line in number_of_lines]

        if len(requested_handles) == 0:
            # If there is no Ref ELEMENT given in the request MESSAGE, then all texts are returned in
            # msg:GetLocalizedTextResponse/msg:Text
            handles = list(self._localized_texts.keys())
        else:
            # If there is at least one Ref ELEMENT given, then msg:GetLocalizedTextResponse/msg:Text contains all texts
            # that match the Ref elements of the msg:GetLocalizedText request MESSAGE.
            handles = requested_handles

        # create a flat list of all localized texts with the requested handles
        texts = []
        for handle in handles:
            with contextlib.suppress(KeyError):
                texts.extend(self._localized_texts[handle])

        # filter languages:
        if requested_langs is not None and len(requested_langs) > 0:
            texts = [t for t in texts if t.Lang in requested_langs]

        # filter requested versions. We need to do it per language, therefore create a lookup with (ref,lang) as key
        tmp_dict = defaultdict(list)
        for text in texts:
            tmp_dict[(text.Ref, text.Lang)].append(text)
        texts = []

        effective_requested_version = requested_version
        if requested_version is None:
            # determine the highest available Version in the storage
            all_versions = []
            for value in self._localized_texts.values():
                all_versions.extend(value)
            if len(all_versions) == 0:
                # ToDo: why return here?
                return []  # there is nothing
            all_versions = [a.Version for a in all_versions if a.Version is not None]
            if len(all_versions) > 0:
                effective_requested_version = max(all_versions)

        # If the referenced text is not available in the specific version, then
        # msg:GetLocalizedTextResponse/msg:Text is empty
        for _, value_list in tmp_dict.items():
            texts.extend([v for v in value_list if v.Version == effective_requested_version])

        # - If there is no NumberOfLines ELEMENT given in the request MESSAGE, then all texts independent of the number
        #   of lines are returned in msg:GetLocalizedTextResponse/msg:Text.
        # - If there is at least one NumberOfLines ELEMENT given, msg:GetLocalizedTextResponse/msg:Text contains texts
        #   that match the number of lines defined by the NumberOfLines elements of the msg:GetLocalizedText request
        #   MESSAGE. Matching in this case means that the number of lines in the text is less or equal to the
        #   NumberOfLines elements.

        if len(i_text_widths) > 0 or len(number_of_lines) > 0:
            if len(number_of_lines) > 0:
                # calculate number of lines for every localized text and add is as member to the object
                for text in texts:
                    text.n_o_l = _calc_number_of_lines(text.text)

            # create again dictionary by ref and language:
            tmp_dict = defaultdict(list)
            for text in texts:
                tmp_dict[(text.Ref, text.Lang)].append(text)
            tmp = []

            if len(i_text_widths) > 0 and len(number_of_lines) > 0:
                # now find for each combination of (width, lines) list the best match
                for value_list in tmp_dict.values():
                    # candidates = []
                    for text_width in i_text_widths:
                        candidates1 = _text_width_filter(value_list,
                                                         text_width)  # returns sorted list of smaller elements
                        for lines_cnt in i_nls:
                            candidates2 = _n_o_l_filter(candidates1, lines_cnt)
                            if len(candidates2) > 0:
                                candidates2.sort(key=lambda obj: obj.TextWidth * obj.n_o_l)  # sort by area size
                                tmp.append(candidates2[-1])  # use the largest one
            elif len(i_text_widths) > 0:
                # filter only text widths
                for value_list in tmp_dict.values():
                    for text_width in i_text_widths:
                        candidates = _text_width_filter(value_list,
                                                        text_width)  # returns sorted list of smaller elements
                        if candidates:
                            tmp.append(candidates[-1])  # use the largest one

            elif len(number_of_lines) > 0:
                # filter only number of lines
                for value_list in tmp_dict.values():
                    for lines_cnt in i_nls:
                        candidates = _n_o_l_filter(value_list, lines_cnt)
                        if candidates:
                            tmp.append(candidates[-1])  # use the largest one
            texts = list(tmp)
        return texts

    def get_supported_languages(self):
        texts = self._flat_list()
        result = set()
        for text in texts:
            result.add(str(text.Lang))
        return list(result)

    def _flat_list(self, ref_list=None):
        if ref_list is None:
            # If there is no Ref ELEMENT given in the request MESSAGE, then all texts are returned in
            # msg:GetLocalizedTextResponse/msg:Text
            handles = list(self._localized_texts.keys())
        else:
            # If there is at least one Ref ELEMENT given, then msg:GetLocalizedTextResponse/msg:Text contains all texts
            # that match the Ref elements of the msg:GetLocalizedText request MESSAGE.
            handles = ref_list

        # create a flat list of all localized texts with the requested handles
        texts = []
        for handle in handles:
            with contextlib.suppress(KeyError):
                texts.extend(self._localized_texts[handle])
        return texts


class LocalizationService(DPWSPortTypeBase):
    port_type_name = PrefixesEnum.SDC.tag('LocalizationService')
    WSDLMessageDescriptions = (WSDLMessageDescription('GetLocalizedText',
                                                      (f'{msg_prefix}:GetLocalizedText',)),
                               WSDLMessageDescription('GetLocalizedTextResponse',
                                                      (f'{msg_prefix}:GetLocalizedTextResponse',)),
                               WSDLMessageDescription('GetSupportedLanguages',
                                                      (f'{msg_prefix}:GetSupportedLanguages',)),
                               WSDLMessageDescription('GetSupportedLanguagesResponse',
                                                      (f'{msg_prefix}:GetSupportedLanguagesResponse',)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('GetLocalizedText', 'literal', 'literal'),
                             WSDLOperationBinding('GetSupportedLanguages', 'literal', 'literal'),)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.localization_storage = LocalizationStorage()

    def register_hosting_service(self, hosting_service):
        super().register_hosting_service(hosting_service)
        actions = self._mdib.sdc_definitions.Actions
        msg_names = self._mdib.sdc_definitions.data_model.msg_names
        hosting_service.register_post_handler(DispatchKey(actions.GetLocalizedText, msg_names.GetLocalizedText),
                                              self._on_get_localized_text)
        hosting_service.register_post_handler(
            DispatchKey(actions.GetSupportedLanguages, msg_names.GetSupportedLanguages),
            self._on_get_supported_languages)

    def _on_get_localized_text(self, request_data):
        data_model = self._sdc_definitions.data_model
        self._logger.debug('_on_get_localized_text')
        cls = data_model.msg_types.GetLocalizedText
        get_localized_text = cls.from_node(request_data.message_data.p_msg.msg_node)
        texts = self.localization_storage.filter_localized_texts(get_localized_text.Ref,
                                                                 get_localized_text.Version,
                                                                 get_localized_text.Lang,
                                                                 get_localized_text.TextWidth,
                                                                 get_localized_text.NumberOfLines)
        response = data_model.msg_types.GetLocalizedTextResponse()
        response.Text.extend(texts)
        response.set_mdib_version_group(self._mdib.mdib_version_group)
        response_envelope = self._sdc_device.msg_factory.mk_reply_soap_message(request_data, response)
        return response_envelope

    def _on_get_supported_languages(self, request_data):
        data_model = self._sdc_definitions.data_model
        self._logger.debug('_on_get_supported_languages')
        languages = self.localization_storage.get_supported_languages()
        response = data_model.msg_types.GetSupportedLanguagesResponse()
        response.Lang.extend(languages)
        response.set_mdib_version_group(self._mdib.mdib_version_group)
        response_envelope = self._sdc_device.msg_factory.mk_reply_soap_message(request_data, response)
        return response_envelope

    def add_wsdl_port_type(self, parent_node):
        port_type = self._mk_port_type_node(parent_node)
        mk_wsdl_two_way_operation(port_type, operation_name='GetLocalizedText')
        mk_wsdl_two_way_operation(port_type, operation_name='GetSupportedLanguages')



"""The module declares the components of a provider.

This serves as dependency injection.
"""
from __future__ import annotations

import copy
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable

from sdc11073.pysoap.msgfactory import MessageFactory
from sdc11073.pysoap.msgreader import MessageReader
from sdc11073.pysoap.soapclient import SoapClient
from sdc11073.pysoap.soapclient_async import SoapClientAsync
from sdc11073.roles.product import DefaultProduct
from sdc11073.roles.waveformprovider.waveformproviderimpl import GenericWaveformProvider

from .operations import get_operation_class
from .porttypes.containmenttreeserviceimpl import ContainmentTreeService
from .porttypes.contextserviceimpl import ContextService
from .porttypes.descriptioneventserviceimpl import DescriptionEventService
from .porttypes.getserviceimpl import GetService
from .porttypes.localizationservice import LocalizationService
from .porttypes.setserviceimpl import SetService
from .porttypes.stateeventserviceimpl import StateEventService
from .porttypes.waveformserviceimpl import WaveformService
from .sco import ScoOperationsRegistry
from .scopesfactory import mk_scopes
from .servicesfactory import mk_all_services
from .subscriptionmgr import PathDispatchingSubscriptionsManager
from .subscriptionmgr_async import SubscriptionsManagerPathAsync

# pylint: disable=cyclic-import
if TYPE_CHECKING:
    from lxml import etree

    from sdc11073 import provider
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.namespaces import PrefixNamespace
    from sdc11073.provider.servicesfactory import HostedServices
    from sdc11073.xml_types.wsd_types import ScopesType

    from .sco import AbstractScoOperationsRegistry
    from .subscriptionmgr_base import SubscriptionManagerProtocol


# pylint: enable=cyclic-import


# Dependency injection: This class defines which component implementations the sdc device will use.
@dataclass()
class SdcProviderComponents:
    """Dependency injection: This class defines which component implementations the sdc provider will use."""

    soap_client_class: type[Any] = None
    msg_factory_class: type[MessageFactory] = None
    msg_reader_class: type[MessageReader] = None
    client_msg_reader_class: type[MessageReader] = None  # the corresponding reader for client
    xml_reader_class: type[MessageReader] = None  # needed to read xml based mdib files
    services_factory: Callable[[provider.SdcProvider, SdcProviderComponents, dict], HostedServices] = None
    operation_cls_getter: Callable[[etree.QName], type] = None
    sco_operations_registry_class: type[AbstractScoOperationsRegistry] = None
    subscriptions_manager_class: dict[str, type[SubscriptionManagerProtocol]] = None
    role_provider_class: type = None
    waveform_provider_class: type | None = None
    scopes_factory: Callable[[ProviderMdibProtocol], ScopesType] = None
    hosted_services: dict = None
    additional_schema_specs: list[PrefixNamespace] = field(default_factory=list)

    def merge(self, other: SdcProviderComponents):
        """Add data from other to self."""

        def _merge(attr_name: str):
            other_value = getattr(other, attr_name)
            if other_value:
                setattr(self, attr_name, other_value)

        _merge('msg_factory_class')
        _merge('msg_reader_class')
        _merge('services_factory')
        _merge('operation_cls_getter')
        _merge('sco_operations_registry_class')
        _merge('role_provider_class')
        _merge('waveform_provider_class')
        _merge('scopes_factory')
        if other.hosted_services is not None:
            self.hosted_services = other.hosted_services
        if other.subscriptions_manager_class is not None:
            for key, value in other.subscriptions_manager_class.items():
                self.subscriptions_manager_class[key] = value
        self.additional_schema_specs = list(set(self.additional_schema_specs).union(set(other.additional_schema_specs)))


default_sdc_provider_components_sync = SdcProviderComponents(
    soap_client_class=SoapClient,
    msg_factory_class=MessageFactory,
    msg_reader_class=MessageReader,
    client_msg_reader_class=MessageReader,
    xml_reader_class=MessageReader,
    services_factory=mk_all_services,
    operation_cls_getter=get_operation_class,
    sco_operations_registry_class=ScoOperationsRegistry,
    subscriptions_manager_class={'StateEvent': PathDispatchingSubscriptionsManager,
                                 'Set': PathDispatchingSubscriptionsManager},
    role_provider_class=DefaultProduct,
    waveform_provider_class=GenericWaveformProvider,
    scopes_factory=mk_scopes,
    # this defines the structure of the services: keys are the names of the dpws hosts,
    # value is a list of port type implementation classes
    hosted_services={'Get': [GetService,
                             LocalizationService],
                     'StateEvent': [StateEventService,
                                    ContextService,
                                    DescriptionEventService,
                                    WaveformService],
                     'Set': [SetService],
                     'ContainmentTree': [ContainmentTreeService]},
)

# async variant
default_sdc_provider_components_async = copy.deepcopy(default_sdc_provider_components_sync)
default_sdc_provider_components_async.soap_client_class = SoapClientAsync
default_sdc_provider_components_async.subscriptions_manager_class = {'StateEvent': SubscriptionsManagerPathAsync,
                                                                   'Set': SubscriptionsManagerPathAsync}

# set default to sync or async variant
default_sdc_provider_components = default_sdc_provider_components_async



"""Schema resolver for SDC11073 XML Schemas."""

from __future__ import annotations

import pathlib
import time
from io import StringIO
from typing import TYPE_CHECKING, Any
from urllib import parse

from lxml import etree

from . import loghelper

if TYPE_CHECKING:
    from .namespaces import NamespaceHelper, PrefixNamespace


def mk_schema_validator(namespaces: list[PrefixNamespace], ns_helper: NamespaceHelper) -> etree.XMLSchema:
    """Create a schema validator."""
    schema_resolver = SchemaResolver(namespaces)
    parser = etree.XMLParser(resolve_entities=True)
    parser.resolvers.add(schema_resolver)
    prefix_enum = ns_helper.prefix_enum
    not_needed = [prefix_enum.XSD]
    # create a schema that includes all used schemas into a single one
    tmp = StringIO()
    tmp.write('<?xml version="1.0" encoding="UTF-8"?>')
    tmp.write(f'<xsd:schema xmlns:xsd="{prefix_enum.XSD.namespace}" elementFormDefault="qualified">\n')
    for entry in namespaces:
        if entry.schema_location_url is not None and entry not in not_needed:
            tmp.write(f'<xsd:import namespace="{entry.namespace}" schemaLocation="{entry.schema_location_url}"/>\n')
    tmp.write('</xsd:schema>')
    all_included = tmp.getvalue().encode('utf-8')

    elem_tree = etree.fromstring(all_included, parser=parser)
    # for unknown reason creating the schema fails sometimes. repeat up to 3 times.
    try:
        return etree.XMLSchema(etree=elem_tree)
    except etree.XMLSchemaParseError:
        time.sleep(0.1)
    try:
        return etree.XMLSchema(etree=elem_tree)
    except etree.XMLSchemaParseError:
        time.sleep(0.5)
    return etree.XMLSchema(etree=elem_tree)


class SchemaResolver(etree.Resolver):
    """A Resolver that uses a list of PrefixNamespace for resolving."""

    def __init__(self, namespaces: list[PrefixNamespace], log_prefix: str | None = None):
        super().__init__()
        self.namespaces = namespaces
        self._logger = loghelper.get_logger_adapter('sdc.schema_resolver', log_prefix)

    def resolve(self, system_url: str, _: Any, context: Any) -> Any:  # return whatever type resolve_string returns
        """Look for xml file location in self.namespaces."""
        self._logger.debug('try to resolve %s', system_url)
        try:
            path = self._get_schema_file_path(system_url)
            if path:
                self._logger.debug('could resolve url %s via lookup to %s', system_url, path)
            else:
                # no lookup, parse url
                parsed = parse.urlparse(system_url)
                path = parsed.path if parsed.scheme == 'file' else system_url
                if path.startswith('/') and path[2] == ':':  # invalid construct like /C:/Temp
                    path = path[1:]
            path = pathlib.Path(path)
            if not path.exists():
                self._logger.error(
                    'no schema file for url "%s": resolved to "%s", but file does not exist',
                    system_url,
                    path,
                )
                return None
            return self.resolve_filename(str(path), context)
        except Exception:
            self._logger.exception('error resolving %s', system_url)
            raise

    def _get_schema_file_path(self, url: str) -> pathlib.Path | None:
        return next((entry.local_schema_file for entry in self.namespaces if entry.schema_location_url == url), None)



from sdc11073.wsdiscovery.service import Service
from sdc11073.xml_types.wsd_types import ScopesType  # make ScopesType visible in wsdiscovery for convenience

from .wsdimpl import (
    WSDiscovery,
    WSDiscoverySingleAdapter,
)

__all__ = ['ScopesType', 'Service', 'WSDiscovery', 'WSDiscoverySingleAdapter']



from sdc11073.dispatch import DispatchKey
from sdc11073.namespaces import PrefixesEnum
from sdc11073.xml_types import msg_qnames
from sdc11073.xml_types.actions import Actions

from .serviceclientbase import HostedServiceClient


class StateEventClient(HostedServiceClient):
    """Client for StateEventService."""

    port_type_name = PrefixesEnum.SDC.tag('StateEventService')
    notifications = (DispatchKey(Actions.EpisodicMetricReport, msg_qnames.EpisodicMetricReport),
                     DispatchKey(Actions.EpisodicAlertReport, msg_qnames.EpisodicAlertReport),
                     DispatchKey(Actions.EpisodicComponentReport, msg_qnames.EpisodicComponentReport),
                     DispatchKey(Actions.EpisodicOperationalStateReport, msg_qnames.EpisodicOperationalStateReport),
                     DispatchKey(Actions.PeriodicMetricReport, msg_qnames.PeriodicMetricReport),
                     DispatchKey(Actions.PeriodicAlertReport, msg_qnames.PeriodicAlertReport),
                     DispatchKey(Actions.PeriodicComponentReport, msg_qnames.PeriodicComponentReport),
                     DispatchKey(Actions.PeriodicOperationalStateReport, msg_qnames.PeriodicOperationalStateReport),
                     DispatchKey(Actions.SystemErrorReport, msg_qnames.SystemErrorReport),
                     )



"""Implementation of audio pause functionality."""
from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073.provider.operations import ExecuteResult
from sdc11073.xml_types.msg_types import InvocationState
from sdc11073.xml_types.pm_types import Coding

from .nomenclature import NomenclatureCodes
from .providerbase import OperationClassGetter, ProviderRole

if TYPE_CHECKING:
    from sdc11073.mdib.descriptorcontainers import AbstractOperationDescriptorProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.provider.operations import ExecuteParameters, OperationDefinitionBase
    from sdc11073.provider.sco import AbstractScoOperationsRegistry

# coded values for SDC audio pause
MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE = Coding(NomenclatureCodes.MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE)
MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE = Coding(NomenclatureCodes.MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE)


class GenericAudioPauseProvider(ProviderRole):
    """Example for handling of global audio pause.

    This provider handles Activate operations with codes "MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE"
    and "MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE".
    Nothing is added to the mdib. If the mdib does not contain these operations, the functionality is not available.
    """

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str):
        super().__init__(mdib, log_prefix)
        self._set_global_audio_pause_operations = []
        self._cancel_global_audio_pause_operations = []

    def make_operation_instance(self,
                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:
        """Create operation handlers for existing mdib entries.

        Handle codes MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE and MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE.
        """
        if operation_descriptor_container.coding == MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE:
            self._logger.debug('instantiating "set audio pause" operation from existing descriptor handle=%s',
                               operation_descriptor_container.Handle)
            set_ap_operation = self._mk_operation_from_operation_descriptor(
                operation_descriptor_container,
                operation_cls_getter,
                operation_handler=self._set_global_audio_pause)
            self._set_global_audio_pause_operations.append(set_ap_operation)
            return set_ap_operation
        if operation_descriptor_container.coding == MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE:
            self._logger.debug('instantiating "cancel audio pause" operation from existing descriptor handle=%s',
                               operation_descriptor_container.Handle)
            cancel_ap_operation = self._mk_operation_from_operation_descriptor(
                operation_descriptor_container,
                operation_cls_getter,
                operation_handler=self._cancel_global_audio_pause)

            self._cancel_global_audio_pause_operations.append(cancel_ap_operation)
            return cancel_ap_operation
        return None

    def _set_global_audio_pause(self, params: ExecuteParameters) -> ExecuteResult:
        """Set global audio pause (ExecuteHandler).

        If global audio pause is initiated, all SystemSignalActivation/State for all alarm systems of the
        product with SystemSignalActivation/Manifestation evaluating to 'Aud' are set to 'Psd'.

        If signal pause is initiated for an alert signal that is not an ACKNOWLEDGE CAPABLE ALERT SIGNAL,
        then the AlertSignalState/ActivationState is set to 'Psd' and the AlertSignalState/Presence to 'Off'.

        If signal pause is initiated for an ACKNOWLEDGEABLE ALERT SIGNAL, the
        AlertSignalState/ActivationState is set to 'Psd' and AlertSignalState/Presence to 'Ack' for that ALERT SIGNAL.
        """
        pm_types = self._mdib.data_model.pm_types
        pm_names = self._mdib.data_model.pm_names

        alert_system_entities = self._mdib.entities.by_node_type(pm_names.AlertSystemDescriptor)
        if len(alert_system_entities) == 0:
            self._logger.warning('_set_global_audio_pause called, but no AlertSystemDescriptor in mdib found')
            return ExecuteResult(params.operation_instance.operation_target_handle, InvocationState.FAILED)

        with self._mdib.alert_state_transaction() as mgr:
            for as_entity in alert_system_entities:
                if as_entity.state.ActivationState != pm_types.AlertActivation.ON:
                    self._logger.info('_set_global_audio_pause: nothing to do for alert system %s', as_entity.handle)
                    continue

                audible_signals = [ssa for ssa in as_entity.state.SystemSignalActivation if
                                   ssa.Manifestation == pm_types.AlertSignalManifestation.AUD]
                active_audible_signals = [ssa for ssa in audible_signals if
                                          ssa.State != pm_types.AlertActivation.PAUSED]
                if len(active_audible_signals) > 0:
                    for ssa in active_audible_signals:
                        ssa.State = pm_types.AlertActivation.PAUSED
                    self._logger.info('SetAudioPauseOperation: set alert system "%s" to paused',
                                      as_entity.handle)

                    # handle all audible alert signals of this alert system
                    all_alert_signal_entities = self._mdib.entities.by_node_type(
                        pm_names.AlertSignalDescriptor)
                    child_alert_signals = [d for d in all_alert_signal_entities if d.parent_handle == as_entity.handle]
                    audible_child_alert_signals = [d for d in child_alert_signals if
                                                   d.descriptor.Manifestation == pm_types.AlertSignalManifestation.AUD]

                    for aud_signal in audible_child_alert_signals:
                        if aud_signal.descriptor.AcknowledgementSupported:
                            if aud_signal.state.ActivationState != pm_types.AlertActivation.PAUSED \
                                    or aud_signal.state.Presence != pm_types.AlertSignalPresence.ACK:
                                aud_signal.state.ActivationState = pm_types.AlertActivation.PAUSED
                                aud_signal.state.Presence = pm_types.AlertSignalPresence.ACK
                                mgr.write_entity(aud_signal)
                        elif aud_signal.state.ActivationState != pm_types.AlertActivation.PAUSED \
                                or aud_signal.state.Presence != pm_types.AlertSignalPresence.OFF:
                            aud_signal.state.ActivationState = pm_types.AlertActivation.PAUSED
                            aud_signal.state.Presence = pm_types.AlertSignalPresence.OFF
                            mgr.write_entity(aud_signal)
                    mgr.write_entity(as_entity)

        return ExecuteResult(params.operation_instance.operation_target_handle,
                             self._mdib.data_model.msg_types.InvocationState.FINISHED)

    def _cancel_global_audio_pause(self, params: ExecuteParameters) -> ExecuteResult:
        """Cancel global audio pause (ExecuteHandler).

        If global audio pause is initiated, all SystemSignalActivation/State for all alarm systems of the product with
        SystemSignalActivation/Manifestation evaluating to 'Aud' shall be set to 'Psd'.
        """
        pm_types = self._mdib.data_model.pm_types
        pm_names = self._mdib.data_model.pm_names
        with self._mdib.alert_state_transaction() as mgr:
            alert_system_entities = self._mdib.entities.by_node_type(pm_names.AlertSystemDescriptor)
            if len(alert_system_entities) == 0:
                self._logger.warning('_cancel_global_audio_pause called, but no AlertSystemDescriptor in mdib found')
                return ExecuteResult(params.operation_instance.operation_target_handle, InvocationState.FAILED)

            for as_entity in alert_system_entities:
                if as_entity.state.ActivationState != pm_types.AlertActivation.ON:
                    self._logger.info('_cancel_global_audio_pause: nothing to do for alert system %s', as_entity.handle)
                    continue

                audible_signals = [ssa for ssa in as_entity.state.SystemSignalActivation if
                                   ssa.Manifestation == pm_types.AlertSignalManifestation.AUD]
                paused_audible_signals = [ssa for ssa in audible_signals if
                                          ssa.State == pm_types.AlertActivation.PAUSED]
                if len(paused_audible_signals) > 0:
                    for ssa in paused_audible_signals:
                        ssa.State = pm_types.AlertActivation.ON
                    self._logger.info('_cancel_global_audio_pause: set alert system "%s" to ON', as_entity.handle)
                    # handle all audible alert signals of this alert system
                    all_alert_signal_entities = self._mdib.entities.by_node_type(pm_names.AlertSignalDescriptor)
                    child_alert_signals = [e for e in all_alert_signal_entities if
                                                   e.parent_handle == as_entity.handle]
                    audible_child_alert_signals = [d for d in child_alert_signals if
                                                   d.descriptor.Manifestation == pm_types.AlertSignalManifestation.AUD]
                    for aud_signal in audible_child_alert_signals:
                        alert_condition_entity = self._mdib.entities.by_handle(aud_signal.descriptor.ConditionSignaled)
                        if alert_condition_entity.state.Presence:
                            # set signal back to 'ON'
                            if aud_signal.state.ActivationState == pm_types.AlertActivation.PAUSED:
                                aud_signal.state.ActivationState = pm_types.AlertActivation.ON
                                aud_signal.state.Presence = pm_types.AlertSignalPresence.ON
                                mgr.write_entity(aud_signal)
                    mgr.write_entity(as_entity)
        return ExecuteResult(params.operation_instance.operation_target_handle,
                             self._mdib.data_model.msg_types.InvocationState.FINISHED)


class AudioPauseProvider(GenericAudioPauseProvider):
    """Handling of global audio pause example.

    This provider guarantees that there are Activate operations with codes "MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE"
    and "MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE". It adds them to mdib if they do not exist.
    """

    def make_missing_operations(self, sco: AbstractScoOperationsRegistry) -> list[OperationDefinitionBase]:
        """Add operations to mdib if they do not exist.

        - code MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE starts alarm pause
        - code MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE cancels alarm pause
        It creates two activate operations with the MDS element as operation target.
        """
        pm_types = self._mdib.data_model.pm_types
        pm_names = self._mdib.data_model.pm_names
        ops = []
        # in this case only the top level sco shall have the additional operations.
        # Check if this is the top level sco (parent is mds)
        parent_entity = self._mdib.entities.by_handle(sco.sco_descriptor_container.parent_handle)
        if pm_names.MdsDescriptor != parent_entity.descriptor.NODETYPE:
            return ops
        operation_cls_getter = sco.operation_cls_getter
        # find mds for this sco
        mds_entity = self._mdib.entities.by_handle(parent_entity.descriptor.source_mds)
        if mds_entity is None:
            msg = f"no source mds found for entity {parent_entity.handle}"
            raise ValueError(msg)

        activate_op_cls = operation_cls_getter(pm_names.ActivateOperationDescriptor)
        if not self._set_global_audio_pause_operations:
            self._logger.debug('adding "set audio pause" operation, no descriptor in mdib (looked for code = %s)',
                               NomenclatureCodes.MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE)
            set_ap_operation = activate_op_cls('AP__ON',
                                               mds_entity.handle,
                                               self._set_global_audio_pause,
                                               coded_value=pm_types.CodedValue(
                                                   NomenclatureCodes.MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE))
            self._set_global_audio_pause_operations.append(set_ap_operation)
            ops.append(set_ap_operation)
        if not self._cancel_global_audio_pause_operations:
            self._logger.debug('adding "cancel audio pause" operation, no descriptor in mdib (looked for code = %s)',
                               NomenclatureCodes.MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE)
            cancel_ap_operation = activate_op_cls('AP__CANCEL',
                                                  mds_entity.handle,
                                                  self._cancel_global_audio_pause,
                                                  coded_value=pm_types.CodedValue(
                                                      NomenclatureCodes.MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE))
            ops.append(cancel_ap_operation)
            self._set_global_audio_pause_operations.append(cancel_ap_operation)
        return ops



"""OperationsManager implements operation handling."""

from __future__ import annotations

import weakref
from collections import deque
from concurrent.futures import Future
from dataclasses import dataclass
from threading import Lock
from typing import TYPE_CHECKING, Protocol

from sdc11073 import loghelper

if TYPE_CHECKING:
    from sdc11073.consumer.manipulator import RequestManipulatorProtocol
    from sdc11073.consumer.serviceclients.serviceclientbase import HostedServiceClient
    from sdc11073.pysoap.msgfactory import CreatedMessage
    from sdc11073.pysoap.msgreader import MessageReader, ReceivedMessage
    from sdc11073.xml_types import msg_types, pm_types


class OperationsManagerProtocol(Protocol):
    """OperationsManager calls an operation.

    It returns a Future object that will contain the result at some point in time.
    """

    def __init__(self, msg_reader: MessageReader, log_prefix: str):
        """Construct the OperationsManager."""
        ...

    def call_operation(
        self,
        hosted_service_client: HostedServiceClient,
        message: CreatedMessage,
        request_manipulator: RequestManipulatorProtocol | None = None,
    ) -> Future:
        """Call an operation.

        An operation call does not return the result of the operation directly. You get a transaction id,
        and will receive the status of this transaction as notification ("OperationInvokedReport").
        This method returns a "Future" object.
        The Future object has a result as soon as a final transaction state is received.
        """

    def on_operation_invoked_report(self, message_data: ReceivedMessage):
        """Check operation state and set future result if it is a final state."""


@dataclass
class OperationResult:
    """OperationResult is the result of a Set operation.

    Usually only the result is relevant, but for testing all intermediate data is also available.
    """

    InvocationInfo: msg_types.InvocationInfo
    InvocationSource: pm_types.InstanceIdentifier | None
    OperationHandleRef: str | None
    OperationTarget: str | None

    set_response: msg_types.AbstractSetResponse
    report_parts: list[
        msg_types.OperationInvokedReportPart
    ]  # contains data of all OperationInvokedReportPart for operation


@dataclass
class OperationData:
    """collect all progress data of a transaction."""

    future_ref: weakref.ref[Future]
    set_response: msg_types.AbstractSetResponse
    report_parts: list[
        msg_types.OperationInvokedReportPart
    ]  # contains data of all OperationInvokedReportPart for operation


class OperationsManager(OperationsManagerProtocol):  # inheriting from protocol to help typing
    """OperationsManager handles the multiple messages that are related to an operation.

    The complex mechanic is hidden from user, he receives the final result.
    """

    def __init__(self, msg_reader: MessageReader, log_prefix: str):
        super().__init__(msg_reader, log_prefix)
        self._msg_reader = msg_reader
        self.log_prefix = log_prefix
        self._logger = loghelper.get_logger_adapter('sdc.client.op_mgr', log_prefix)
        self._transactions: dict[int, OperationData] = {}
        self._transactions_lock = Lock()
        # An OperationInvokedReport can be received even before the response of the set operation is received.
        # This means we must always store the last n OperationInvokedReportParts, one of them might already be the
        # needed one.
        self._last_operation_invoked_reports: deque[msg_types.OperationInvokedReportPart] = deque(maxlen=50)
        msg_types = msg_reader.msg_types
        self.nonFinalOperationStates = (msg_types.InvocationState.WAIT, msg_types.InvocationState.START)

    def call_operation(
        self,
        hosted_service_client: HostedServiceClient,
        message: CreatedMessage,
        request_manipulator: RequestManipulatorProtocol | None = None,
    ) -> Future:
        """Call an operation."""
        future_object = Future()
        message_data = hosted_service_client.post_message(
            message,
            msg='call Operation',
            request_manipulator=request_manipulator,
        )
        with self._transactions_lock:
            msg_types = self._msg_reader.msg_types
            abstract_set_response = msg_types.AbstractSetResponse.from_node(message_data.p_msg.msg_node)
            invocation_info = abstract_set_response.InvocationInfo
            if invocation_info.InvocationState in (
                msg_types.InvocationState.FAILED,
                msg_types.InvocationState.CANCELLED,
                msg_types.InvocationState.CANCELLED_MANUALLY,
            ):
                # do not wait for an OperationInvokedReport
                operation_result = OperationResult(
                    abstract_set_response.InvocationInfo,
                    None,
                    None,
                    None,
                    abstract_set_response,
                    [],
                )
                future_object.set_result(operation_result)
                return future_object
            transaction_id = invocation_info.TransactionId
            # now look for all related report parts and add them to result
            parts = [
                part
                for part in self._last_operation_invoked_reports
                if part.InvocationInfo.TransactionId == transaction_id
            ]
            # now look for a final report part
            final_parts = [
                part for part in parts if part.InvocationInfo.InvocationState not in self.nonFinalOperationStates
            ]
            if final_parts:
                report_part = final_parts[0]  # assuming there is only one
                future_object.set_result(self._mk_operation_result(report_part, abstract_set_response, parts))
            else:
                self._logger.info(  # noqa: PLE1205
                    'call_operation: transaction_id {} registered, state={}',
                    invocation_info.TransactionId,
                    invocation_info.InvocationState,
                )
                self._transactions[transaction_id] = OperationData(
                    weakref.ref(future_object),
                    abstract_set_response,
                    parts,
                )
        return future_object

    def on_operation_invoked_report(self, message_data: ReceivedMessage):
        """Check operation state and set future result if it is a final state."""
        msg_types = self._msg_reader.msg_types
        operation_invoked_report = msg_types.OperationInvokedReport.from_node(message_data.p_msg.msg_node)

        with self._transactions_lock:
            for report_part in operation_invoked_report.ReportPart:
                invocation_state = report_part.InvocationInfo.InvocationState
                transaction_id = report_part.InvocationInfo.TransactionId
                self._logger.debug(  # noqa: PLE1205
                    '{}on_operation_invoked_report: got transaction_id {} state {}',
                    self.log_prefix,
                    transaction_id,
                    invocation_state,
                )
                if transaction_id in self._transactions:
                    self._transactions[transaction_id].report_parts.append(report_part)
                    if invocation_state in self.nonFinalOperationStates:
                        pass
                    else:
                        operation_data = self._transactions.pop(transaction_id, None)
                        future_object = operation_data.future_ref()
                        if future_object is None:
                            # client gave up.
                            self._logger.debug('transaction_id {} given up', transaction_id)  # noqa: PLE1205
                        else:
                            future_object.set_result(
                                self._mk_operation_result(
                                    report_part,
                                    operation_data.set_response,
                                    operation_data.report_parts,
                                ),
                            )
                else:
                    self._last_operation_invoked_reports.append(report_part)

    def _mk_operation_result(
        self,
        current_report_part: msg_types.OperationInvokedReportPart,
        set_response: msg_types.AbstractSetResponse,
        all_report_parts: list[msg_types.OperationInvokedReportPart],
    ) -> OperationResult:
        return OperationResult(
            current_report_part.InvocationInfo,
            current_report_part.InvocationSource,
            current_report_part.OperationHandleRef,
            current_report_part.OperationTarget,
            set_response,
            all_report_parts,
        )



"""A helper for xml name space handling."""

from __future__ import annotations

import pathlib
from enum import Enum
from typing import NamedTuple

from lxml import etree

from sdc11073 import xml_utils


class PrefixNamespace(NamedTuple):
    """Represents a namespace with an optional prefix and schema location."""

    prefix: str
    namespace: str
    schema_location_url: str | None
    local_schema_file: pathlib.Path | None

    def tag(self, localname: str) -> xml_utils.QName:
        """Create a QName for the given localname."""
        return xml_utils.QName(self.namespace, localname)

    def doc_name(self, localname: str) -> str:
        """Get the string representation with an optional prefix."""
        if self.prefix:
            return f'{self.prefix}:{localname}'
        return localname


schema_folder = pathlib.Path(__file__).parent.joinpath('xsd')


class PrefixesEnum(PrefixNamespace, Enum):
    """Enum for various XML namespaces with prefixes."""

    MSG = PrefixNamespace(
        'msg',
        'http://standards.ieee.org/downloads/11073/11073-10207-2017/message',
        'http://standards.ieee.org/downloads/11073/11073-10207-2017/BICEPS_MessageModel.xsd',
        schema_folder.joinpath('BICEPS_MessageModel.xsd'),
    )
    PM = PrefixNamespace(
        'dom',
        'http://standards.ieee.org/downloads/11073/11073-10207-2017/participant',
        'http://standards.ieee.org/downloads/11073/11073-10207-2017/BICEPS_ParticipantModel.xsd',
        schema_folder.joinpath('BICEPS_ParticipantModel.xsd'),
    )
    EXT = PrefixNamespace(
        'ext',
        'http://standards.ieee.org/downloads/11073/11073-10207-2017/extension',
        'http://standards.ieee.org/downloads/11073/11073-10207-2017/ExtensionPoint.xsd',
        schema_folder.joinpath('ExtensionPoint.xsd'),
    )
    MDPWS = PrefixNamespace('mdpws', 'http://standards.ieee.org/downloads/11073/11073-20702-2016', None, None)
    SDC = PrefixNamespace('sdc', 'http://standards.ieee.org/downloads/11073/11073-20701-2018', None, None)
    WSE = PrefixNamespace(
        'wse',
        'http://schemas.xmlsoap.org/ws/2004/08/eventing',
        'http://schemas.xmlsoap.org/ws/2004/08/eventing',
        schema_folder.joinpath('eventing.xsd'),
    )
    XSD = PrefixNamespace('xsd', 'http://www.w3.org/2001/XMLSchema', None, None)
    XSI = PrefixNamespace('xsi', 'http://www.w3.org/2001/XMLSchema-instance', None, None)
    WSA = PrefixNamespace(
        'wsa',
        'http://www.w3.org/2005/08/addressing',
        'http://www.w3.org/2006/03/addressing/ws-addr.xsd',
        schema_folder.joinpath('ws-addr.xsd'),
    )
    WSX = PrefixNamespace(
        'wsx',  # Meta Data Exchange
        'http://schemas.xmlsoap.org/ws/2004/09/mex',
        'http://schemas.xmlsoap.org/ws/2004/09/mex',
        schema_folder.joinpath('MetadataExchange.xsd'),
    )
    DPWS = PrefixNamespace(
        'dpws',
        'http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01',
        'http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01',
        schema_folder.joinpath('wsdd-dpws-1.1-schema-os.xsd'),
    )
    WSD = PrefixNamespace(
        'wsd',
        'http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01',
        'http://docs.oasis-open.org/ws-dd/discovery/1.1/os/wsdd-discovery-1.1-schema-os.xsd',
        schema_folder.joinpath('wsdd-discovery-1.1-schema-os.xsd'),
    )
    S12 = PrefixNamespace(
        's12',
        'http://www.w3.org/2003/05/soap-envelope',
        'http://www.w3.org/2003/05/soap-envelope',
        schema_folder.joinpath('soap-envelope.xsd'),
    )
    XML = PrefixNamespace(
        'xml',
        'http://www.w3.org/XML/1998/namespace',
        'http://www.w3.org/2001/xml.xsd',
        schema_folder.joinpath('xml.xsd'),
    )
    WXF = PrefixNamespace(
        'wxf',  # ws-transfer
        'http://schemas.xmlsoap.org/ws/2004/09/transfer',
        None,
        None,
    )
    WSDL = PrefixNamespace(
        'wsdl',
        'http://schemas.xmlsoap.org/wsdl/',
        'http://schemas.xmlsoap.org/wsdl/',
        schema_folder.joinpath('wsdl.xsd'),
    )
    WSDL12 = PrefixNamespace(
        'wsdl12',
        'http://schemas.xmlsoap.org/wsdl/soap12/',
        None,
        None,
    )  # old soap 12 namespace, used in wsdl 1.1. only for wsdl
    WSP = PrefixNamespace('wsp', 'http://www.w3.org/ns/ws-policy', None, None)


class NamespaceHelper:
    """A helper class for managing XML namespaces and their prefixes."""

    def __init__(self, prefixes_enum: type[PrefixesEnum], default_ns: str | None = None):
        self.prefix_enum = prefixes_enum
        self._lookup = {}
        for enum_item in prefixes_enum:
            self._lookup[enum_item.name] = enum_item.value

        self._default_ns = default_ns

        self._prefix_map = {x.namespace: x.prefix for x in self._lookup.values()}  # map namespace to prefix
        self.ns_map = {x.prefix: x.namespace for x in self._lookup.values()}  # map prefix to namespace

    @property
    def MSG(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['MSG']

    @property
    def PM(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['PM']

    @property
    def EXT(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['EXT']

    @property
    def SDC(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['SDC']

    @property
    def WSE(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['WSE']

    @property
    def XSI(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['XSI']

    @property
    def WSA(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['WSA']

    @property
    def WSX(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['WSX']

    @property
    def DPWS(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['DPWS']

    @property
    def MDPWS(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['MDPWS']

    @property
    def WSD(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['WSD']

    @property
    def S12(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['S12']

    @property
    def XML(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['XML']

    @property
    def WSDL(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['WSDL']

    @property
    def WSDL12(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['WSDL12']

    @property
    def WSP(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['WSP']

    @property
    def WXF(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['WXF']

    @property
    def XSD(self) -> PrefixNamespace:  # noqa: D102, N802
        return self._lookup['XSD']

    def partial_map(self, *prefix: PrefixNamespace) -> dict:
        """Get a dictionary with prefix as key, namespace as value."""
        ret = {}
        for p in prefix:
            if p.namespace == self._default_ns:
                ret[None] = p.namespace
            ret[p.prefix] = p.namespace
        return ret

    def doc_name_from_qname(self, qname: etree.QName) -> str:
        """Get the prefix:name string, or only name (if default namespace is used)."""
        if qname.namespace is not None and qname.namespace == self._default_ns:
            return qname.localname
        prefix = self._prefix_map[qname.namespace]
        return f'{prefix}:{qname.localname}'

    def text_to_qname(self, text: str, nsmap: dict[str, str] | None = None) -> etree.QName:
        """Convert a text to a qname."""
        ns_map = nsmap or self.ns_map
        return text_to_qname(text, ns_map)


default_ns_helper = NamespaceHelper(PrefixesEnum)

# some constants from ws-addressing
WSA_ANONYMOUS = PrefixesEnum.WSA.namespace + '/anonymous'
WSA_NONE = PrefixesEnum.WSA.namespace + '/none'


def docname_from_qname(qname: etree.QName, ns_map: dict) -> str:
    """Get a prefix:name string, or only name (if default namespace is used)."""
    prefixmap = {v: k for k, v in ns_map.items()}
    prefix = prefixmap.get(qname.namespace)
    if prefix is None:
        return qname.localname
    return f'{prefix}:{qname.localname}'


def text_to_qname(text: str, doc_nsmap: dict[str, str]) -> xml_utils.QName:
    """Convert a text to a qname."""
    elements = text.split(':')
    prefix = None if len(elements) == 1 else elements[0]
    name = elements[-1]
    try:
        return xml_utils.QName(doc_nsmap[prefix], name)
    except KeyError as ex:
        raise KeyError(f'Cannot make QName for {text}, prefix is not in nsmap: {doc_nsmap.keys()}') from ex  # noqa: EM102


QN_TYPE = etree.QName(PrefixesEnum.XSI.namespace, 'type')  # frequently used QName, central definition


class EventingActions:
    """Constants for eventing actions."""

    Subscribe = PrefixesEnum.WSE.namespace + '/Subscribe'
    SubscribeResponse = PrefixesEnum.WSE.namespace + '/SubscribeResponse'
    SubscriptionEnd = PrefixesEnum.WSE.namespace + '/SubscriptionEnd'
    Unsubscribe = PrefixesEnum.WSE.namespace + '/Unsubscribe'
    UnsubscribeResponse = PrefixesEnum.WSE.namespace + '/UnsubscribeResponse'
    Renew = PrefixesEnum.WSE.namespace + '/Renew'
    RenewResponse = PrefixesEnum.WSE.namespace + '/RenewResponse'
    GetStatus = PrefixesEnum.WSE.namespace + '/GetStatus'
    GetStatusResponse = PrefixesEnum.WSE.namespace + '/GetStatusResponse'



from . import xml_structure as struct
from .basetypes import XMLTypeBase, MessageType, ElementWithTextList
from .addressing_types import EndpointReferenceType
from sdc11073.namespaces import default_ns_helper
from typing import Optional

########## Meta Data Exchange #########
wsa_tag = default_ns_helper.WSA.tag  # shortcut
wsd_tag = default_ns_helper.WSD.tag  # shortcut

class QNameListType(struct.NodeTextQNameListProperty):
    pass

class ScopesType(ElementWithTextList):
    # text is a URI list
    MatchBy = struct.AnyURIAttributeProperty('MatchBy')
    _props = ('MatchBy',)

    def __init__(self, value: Optional[str] = None, match_by: Optional[str] = None):
        super().__init__()
        if value is not None:
            self.text.append(value)
        self.MatchBy = match_by


class HelloType(MessageType):
    NODETYPE = wsd_tag('Hello')
    action = f'{default_ns_helper.WSD.namespace}/Hello'
    EndpointReference = struct.SubElementProperty(wsa_tag('EndpointReference'),
                                                  value_class=EndpointReferenceType,
                                                  default_py_value=EndpointReferenceType())
    Types  = QNameListType(wsd_tag('Types'), is_optional=True)
    Scopes  = struct.SubElementProperty(wsd_tag('Scopes'),
                                        value_class=ScopesType,
                                        is_optional=True)
    XAddrs  = struct.NodeTextListProperty(wsd_tag('XAddrs'),
                                          value_class=str,
                                          is_optional=True)
    MetadataVersion = struct.NodeIntProperty(wsd_tag('MetadataVersion'), default_py_value=1)
    _props = ('EndpointReference', 'Types', 'Scopes', 'XAddrs', 'MetadataVersion')


class ByeType(MessageType):
    NODETYPE = wsd_tag('Bye')
    action = f'{default_ns_helper.WSD.namespace}/Bye'
    EndpointReference = struct.SubElementProperty(wsa_tag('EndpointReference'),
                                                  value_class=EndpointReferenceType,
                                                  default_py_value=EndpointReferenceType())
    Types  = QNameListType(wsd_tag('Types'), is_optional=True)
    Scopes  = struct.SubElementProperty(wsd_tag('Scopes'),
                                        value_class=ScopesType,
                                        is_optional=True)
    XAddrs  = struct.NodeTextListProperty(wsd_tag('XAddrs'),
                                          value_class=str,
                                          is_optional=True)
    MetadataVersion = struct.NodeIntProperty(wsd_tag('MetadataVersion'), is_optional=True)
    _props = ('EndpointReference', 'Types', 'Scopes', 'XAddrs', 'MetadataVersion')


class ProbeType(MessageType):
    NODETYPE = wsd_tag('Probe')
    action = f'{default_ns_helper.WSD.namespace}/Probe'
    Types  = QNameListType(wsd_tag('Types'), is_optional=True)
    Scopes  = struct.SubElementProperty(wsd_tag('Scopes'),
                                        value_class=ScopesType,
                                        is_optional=True)
    _props = ('Types', 'Scopes')


class ProbeMatchType(XMLTypeBase):
    EndpointReference = struct.SubElementProperty(wsa_tag('EndpointReference'),
                                                  value_class=EndpointReferenceType,
                                                  default_py_value=EndpointReferenceType())
    Types  = QNameListType(wsd_tag('Types'), is_optional=True)
    Scopes  = struct.SubElementProperty(wsd_tag('Scopes'),
                                        value_class=ScopesType,
                                        is_optional=True)
    XAddrs  = struct.NodeTextListProperty(wsd_tag('XAddrs'),
                                          value_class=str,
                                          is_optional=True)
    MetadataVersion = struct.NodeIntProperty(wsd_tag('MetadataVersion'), default_py_value=1)
    _props = ('EndpointReference', 'Types', 'Scopes', 'XAddrs', 'MetadataVersion')


class ProbeMatchesType(MessageType):
    NODETYPE = wsd_tag('ProbeMatches')
    action = f'{default_ns_helper.WSD.namespace}/ProbeMatches'
    ProbeMatch = struct.SubElementListProperty(wsd_tag('ProbeMatch'),
                                               value_class=ProbeMatchType)
    _props = ('ProbeMatch',)
    additional_namespaces = (default_ns_helper.WSD, default_ns_helper.WSA)


class ResolveType(MessageType):
    NODETYPE = wsd_tag('Resolve')
    action = f'{default_ns_helper.WSD.namespace}/Resolve'
    EndpointReference = struct.SubElementProperty(wsa_tag('EndpointReference'),
                                                  value_class=EndpointReferenceType,
                                                  default_py_value=EndpointReferenceType())
    _props = ('EndpointReference',)


class ResolveMatchType(XMLTypeBase):
    EndpointReference = struct.SubElementProperty(wsa_tag('EndpointReference'),
                                                  value_class=EndpointReferenceType,
                                                  default_py_value=EndpointReferenceType())
    Types  = QNameListType(wsd_tag('Types'), is_optional=True)
    Scopes  = struct.SubElementProperty(wsd_tag('Scopes'),
                                        value_class=ScopesType,
                                        is_optional=True)
    XAddrs  = struct.NodeTextListProperty(wsd_tag('XAddrs'),
                                          value_class=str,
                                          is_optional=True)
    MetadataVersion = struct.NodeIntProperty(wsd_tag('MetadataVersion'), default_py_value=1)
    _props = ('EndpointReference', 'Types', 'Scopes', 'XAddrs', 'MetadataVersion')


class ResolveMatchesType(MessageType):
    NODETYPE = wsd_tag('ResolveMatches')
    action = f'{default_ns_helper.WSD.namespace}/ResolveMatches'
    ResolveMatch = struct.SubElementProperty(wsd_tag('ResolveMatch'),
                                               value_class=ResolveMatchType,
                                             is_optional=True)
    _props = ('ResolveMatch',)


class AppSequenceType(XMLTypeBase):
    NODETYPE = wsd_tag('AppSequence')
    # used in soap header
    InstanceId = struct.IntegerAttributeProperty('InstanceId', is_optional=False)
    SequenceId = struct.AnyURIAttributeProperty('SequenceId')
    MessageNumber = struct.IntegerAttributeProperty('MessageNumber', is_optional=False)
    _props = ('InstanceId', 'SequenceId', 'MessageNumber')



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol, runtime_checkable

from sdc11073.dispatch import DispatchKey
from sdc11073.namespaces import PrefixesEnum

from .porttypebase import (
    ServiceWithOperations,
    WSDLMessageDescription,
    WSDLOperationBinding,
    mk_wsdl_one_way_operation,
    mk_wsdl_two_way_operation,
    msg_prefix,
)

if TYPE_CHECKING:
    from enum import Enum

    from sdc11073.provider.sco import OperationDefinition
    from sdc11073.mdib.mdibbase import MdibVersionGroup


@runtime_checkable
class SetServiceProtocol(Protocol):
    def notify_operation(self,
                         operation: OperationDefinition,
                         transaction_id: int,
                         invocation_state: Enum,
                         mdib_version_group: MdibVersionGroup,
                         operation_target: str | None = None,
                         error: Enum | None = None,
                         error_message: str | None = None):
        ...


class SetService(ServiceWithOperations):
    port_type_name = PrefixesEnum.SDC.tag('SetService')
    WSDLMessageDescriptions = (WSDLMessageDescription('Activate',
                                                      (f'{msg_prefix}:Activate',)),
                               WSDLMessageDescription('ActivateResponse',
                                                      (f'{msg_prefix}:ActivateResponse',)),
                               WSDLMessageDescription('SetString',
                                                      (f'{msg_prefix}:SetString',)),
                               WSDLMessageDescription('SetStringResponse',
                                                      (f'{msg_prefix}:SetStringResponse',)),
                               WSDLMessageDescription('SetComponentState',
                                                      (f'{msg_prefix}:SetComponentState',)),
                               WSDLMessageDescription('SetComponentStateResponse',
                                                      (f'{msg_prefix}:SetComponentStateResponse',)),
                               WSDLMessageDescription('SetAlertState',
                                                      (f'{msg_prefix}:SetAlertState',)),
                               WSDLMessageDescription('SetAlertStateResponse',
                                                      (f'{msg_prefix}:SetAlertStateResponse',)),
                               WSDLMessageDescription('SetMetricState',
                                                      (f'{msg_prefix}:SetMetricState',)),
                               WSDLMessageDescription('SetMetricStateResponse',
                                                      (f'{msg_prefix}:SetMetricStateResponse',)),
                               WSDLMessageDescription('SetValue',
                                                      (f'{msg_prefix}:SetValue',)),
                               WSDLMessageDescription('SetValueResponse',
                                                      (f'{msg_prefix}:SetValueResponse',)),
                               WSDLMessageDescription('OperationInvokedReport',
                                                      (f'{msg_prefix}:OperationInvokedReport',)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('Activate', 'literal', 'literal'),  # fault?
                             WSDLOperationBinding('SetString', 'literal', 'literal'),  # fault?
                             WSDLOperationBinding('SetComponentState', 'literal', 'literal'),  # fault?
                             WSDLOperationBinding('SetAlertState', 'literal', 'literal'),  # fault?
                             WSDLOperationBinding('SetMetricState', 'literal', 'literal'),  # fault?
                             WSDLOperationBinding('SetValue', 'literal', 'literal'),  # fault?
                             WSDLOperationBinding('OperationInvokedReport', None, 'literal'),
                             )

    def register_hosting_service(self, hosting_service):
        super().register_hosting_service(hosting_service)
        actions = self._mdib.sdc_definitions.Actions
        msg_names = self._mdib.sdc_definitions.data_model.msg_names
        hosting_service.register_post_handler(DispatchKey(actions.Activate, msg_names.Activate),
                                              self._on_activate)
        hosting_service.register_post_handler(DispatchKey(actions.SetValue, msg_names.SetValue),
                                              self._on_set_value)
        hosting_service.register_post_handler(DispatchKey(actions.SetString, msg_names.SetString),
                                              self._on_set_string)
        hosting_service.register_post_handler(DispatchKey(actions.SetMetricState, msg_names.SetMetricState),
                                              self._on_set_metric_state)
        hosting_service.register_post_handler(DispatchKey(actions.SetAlertState, msg_names.SetAlertState),
                                              self._on_set_alert_state)
        hosting_service.register_post_handler(DispatchKey(actions.SetComponentState, msg_names.SetComponentState),
                                              self._on_set_component_state)

    def _on_activate(self, request_data):
        """Handler for Active calls.
        It enqueues an operation and generates the expected operation invoked report.
        """
        data_model = self._sdc_definitions.data_model
        msg_node = request_data.message_data.p_msg.msg_node
        activate = data_model.msg_types.Activate.from_node(msg_node)
        # ToDo: convert arguments to specific python types
        response = data_model.msg_types.ActivateResponse()
        return self._handle_operation_request(request_data, activate, response)

    def _on_set_value(self, request_data):
        """Handler for SetValue calls.
        It enqueues an operation and generates the expected operation invoked report.
        """
        data_model = self._sdc_definitions.data_model
        self._logger.debug('_on_set_value')
        msg_node = request_data.message_data.p_msg.msg_node
        set_value = data_model.msg_types.SetValue.from_node(msg_node)
        response = data_model.msg_types.SetValueResponse()
        return self._handle_operation_request(request_data, set_value, response)

    def _on_set_string(self, request_data):
        """Handler for SetString calls.
        It enqueues an operation and generates the expected operation invoked report.
        """
        data_model = self._sdc_definitions.data_model
        self._logger.debug('_on_set_string')
        msg_node = request_data.message_data.p_msg.msg_node
        set_string = data_model.msg_types.SetString.from_node(msg_node)
        response = data_model.msg_types.SetStringResponse()
        return self._handle_operation_request(request_data, set_string, response)

    def _on_set_metric_state(self, request_data):
        """Handler for SetMetricState calls.
        It enqueues an operation and generates the expected operation invoked report.
        """
        data_model = self._sdc_definitions.data_model
        self._logger.debug('_on_set_metric_state')
        msg_node = request_data.message_data.p_msg.msg_node
        set_metric_state = data_model.msg_types.SetMetricState.from_node(msg_node)
        response = data_model.msg_types.SetMetricStateResponse()
        return self._handle_operation_request(request_data, set_metric_state, response)

    def _on_set_alert_state(self, request_data):
        """Handler for SetMetricState calls.
        It enqueues an operation and generates the expected operation invoked report.
        """
        data_model = self._sdc_definitions.data_model
        self._logger.debug('_on_set_alert_state')
        msg_node = request_data.message_data.p_msg.msg_node
        set_alert_state = data_model.msg_types.SetAlertState.from_node(msg_node)
        response = data_model.msg_types.SetAlertStateResponse()
        return self._handle_operation_request(request_data, set_alert_state, response)

    def _on_set_component_state(self, request_data):
        """Handler for SetComponentState calls.
        It enqueues an operation and generates the expected operation invoked report.
        """
        data_model = self._sdc_definitions.data_model
        self._logger.debug('_on_set_component_state')
        msg_node = request_data.message_data.p_msg.msg_node
        set_component_state = data_model.msg_types.SetComponentState.from_node(msg_node)
        response = data_model.msg_types.SetComponentStateResponse()
        return self._handle_operation_request(request_data, set_component_state, response)

    def notify_operation(self,
                         operation: OperationDefinition,
                         transaction_id: int,
                         invocation_state: Enum,
                         mdib_version_group: MdibVersionGroup,
                         operation_target: str | None = None,
                         error: Enum | None = None,
                         error_message: str | None = None):
        data_model = self._sdc_definitions.data_model
        nsh = data_model.ns_helper
        operation_handle_ref = operation.handle
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.OperationInvokedReport()
        report.set_mdib_version_group(mdib_version_group)
        report_part = report.add_report_part()
        report_part.InvocationInfo.TransactionId = transaction_id
        report_part.InvocationInfo.InvocationState = invocation_state
        if error is not None:
            report_part.InvocationInfo.InvocationError = error
        if error_message is not None:
            report_part.InvocationInfo.InvocationErrorMessage.append(data_model.pm_types.LocalizedText(error_message))
        # implemented is only SDC R0077 for value of invocationSource:
        # Extension = "AnonymousSdcParticipant".
        # a known participant (R0078) is currently not supported
        # ToDo: implement R0078
        report_part.InvocationSource = data_model.pm_types.InstanceIdentifier(
            nsh.SDC.namespace, extension_string='AnonymousSdcParticipant')
        report_part.OperationHandleRef = operation_handle_ref
        report_part.OperationTarget = operation_target
        ns_map = nsh.partial_map(nsh.PM, nsh.MSG, nsh.XSI, nsh.EXT, nsh.XML)
        body_node = report.as_etree_node(report.NODETYPE, ns_map)
        self._logger.info(
            'notify_operation transaction={} operation_handle_ref={}, operationState={}, error={}, errorMessage={}',
            transaction_id, operation_handle_ref, invocation_state, error, error_message)
        subscription_mgr.send_to_subscribers(body_node, report.action.value, mdib_version_group)

    def handled_actions(self) -> List[str]:
        return [self._sdc_device.sdc_definitions.Actions.OperationInvokedReport]

    def add_wsdl_port_type(self, parent_node):
        port_type = self._mk_port_type_node(parent_node, True)
        mk_wsdl_two_way_operation(port_type, operation_name='Activate')
        mk_wsdl_two_way_operation(port_type, operation_name='SetString')
        mk_wsdl_two_way_operation(port_type, operation_name='SetComponentState')
        mk_wsdl_two_way_operation(port_type, operation_name='SetAlertState')
        mk_wsdl_two_way_operation(port_type, operation_name='SetMetricState')
        mk_wsdl_two_way_operation(port_type, operation_name='SetValue')
        mk_wsdl_one_way_operation(port_type, operation_name='OperationInvokedReport')



"""HTTP server implementation with threading and request dispatching."""

from __future__ import annotations

import logging
import socket
import socketserver
import threading
from dataclasses import dataclass
from typing import TYPE_CHECKING

from sdc11073.dispatch import PathElementRegistry
from sdc11073.loghelper import LoggerAdapter

from .httprequesthandler import DispatchingRequestHandler

if TYPE_CHECKING:
    from collections.abc import Iterable

    from sdc11073 import certloader


@dataclass(frozen=True)
class _ThreadInfo:
    thread: threading.Thread
    request: socket.socket
    client_address: tuple


class _ThreadingHTTPServer(socketserver.TCPServer):
    """Each request is handled in a thread."""

    def __init__(
        self,
        logger: LoggerAdapter,
        server_address: tuple[str, int],
        chunk_size: int,
        supported_encodings: Iterable[str],
    ):
        self.daemon_threads = True
        self.threads = []
        self.logger = logger
        self.dispatcher = PathElementRegistry()
        self.chunk_size = chunk_size
        self.supported_encodings = supported_encodings
        self.server_port: int | None = None
        super().__init__(server_address, DispatchingRequestHandler)

    def server_bind(self):
        """Override server_bind to store the server name."""
        socketserver.TCPServer.server_bind(self)
        self.server_port = self.server_address[1]

    def process_request_thread(self, request, client_address):  # noqa: ANN001
        """Same as in BaseServer but as a thread."""  # noqa: D401
        try:
            self.finish_request(request, client_address)
        except (ConnectionResetError, ConnectionAbortedError) as ex:
            self.logger.info('Connection reset by %s: %s', client_address, ex)
        except Exception:  # noqa: BLE001
            self.handle_error(request, client_address)
        finally:
            self.shutdown_request(request)
            # this thread will close after return from this method, it can already be removed from self.threads
            for thread_info in self.threads:
                if thread_info.request == request:
                    self.threads.remove(thread_info)

    def process_request(self, request, client_address):  # noqa: ANN001
        """Start a new thread to process the request."""
        thread = threading.Thread(
            target=self.process_request_thread,
            args=(request, client_address),
            name=f'SubscrRecv{client_address}',
        )
        thread.daemon = True
        self.threads.append(_ThreadInfo(thread, request, client_address))
        thread.start()

    def server_close(self):
        super().server_close()
        if self.dispatcher is not None:
            self.dispatcher.methods = {}
            self.dispatcher = None  # this leads to a '503' reaction in SOAPNotificationsHandler
        for thread_info in self.threads:
            if thread_info.thread.is_alive():
                try:
                    thread_info.request.shutdown(socket.SHUT_RDWR)
                    thread_info.request.close()
                    self.logger.info('closed socket for notifications from %s', thread_info.client_address)
                except OSError:
                    # the connection is already closed
                    continue
                except Exception as ex:  # noqa: BLE001
                    self.logger.warning(
                        'error closing socket for notifications from %s: %s',
                        thread_info.client_address,
                        ex,
                    )


class HttpServerThreadBase(threading.Thread):
    """A Thread running a ThreadingHTTPServer."""

    def __init__(
        self,
        my_ipaddress: str,
        ssl_context: certloader.SSLContextContainer | None,
        supported_encodings: Iterable[str],
        logger: logging.Logger | LoggerAdapter,
        chunk_size: int = 0,
    ):
        """Run a ThreadingHTTPServer in a thread, so that it can be stopped without blocking.

        Handling of requests happens in two stages:
        - the http server instantiates a request handler with the request
        - the request handler forwards the handling itself to a dispatcher (due to the dynamic nature of the handling).
        :param my_ipaddress: The ip address that the http server shall bind to (no port!)
        :param ssl_context: a ssl.SslContext instance or None
        :param supported_encodings: a list of strings
        :param logger: a python logger
        :param chunk_size: if value > 0, messages are split into chunks of this size.
        """
        super().__init__(name='Dev_SdcHttpServerThread')
        self.daemon = True

        self._my_ipaddress = my_ipaddress
        self._ssl_context = ssl_context
        self.my_port = None
        self.httpd = None
        self.supported_encodings = supported_encodings
        if isinstance(logger, logging.Logger):
            self.logger = LoggerAdapter(logger)
        else:
            self.logger = logger
        self.chunk_size = chunk_size
        # create and set up the dispatcher for all incoming requests
        self.started_evt = threading.Event()  # helps to wait until thread has initialized is variables
        self._stop_requested = False
        self.base_url = None

    def run(self):
        """Run the http server."""
        self._stop_requested = False
        try:
            myport = 0  # zero means that OS selects a free port
            self.httpd = _ThreadingHTTPServer(
                self.logger,
                (self._my_ipaddress, myport),
                self.chunk_size,
                self.supported_encodings,
            )
            self.my_port = self.httpd.server_port
            self.logger.info('starting http server on %s:%s', self._my_ipaddress, self.my_port)
            if self._ssl_context:
                self.httpd.socket = self._ssl_context.wrap_socket(self.httpd.socket, server_side=True)
                self.base_url = f'https://{self._my_ipaddress}:{self.my_port}/'
            else:
                self.base_url = f'http://{self._my_ipaddress}:{self.my_port}/'

            self.started_evt.set()
            self.httpd.serve_forever()
        except Exception:
            if not self._stop_requested:
                self.logger.exception('Unhandled Exception at thread runtime. Thread will abort!')
            raise
        finally:
            self.logger.info('http server stopped.')

    @property
    def dispatcher(self) -> PathElementRegistry:
        """Return the dispatcher responsible for handling requests."""
        if not self.started_evt.is_set():
            raise RuntimeError('http server not started yet, dispatcher not available')
        return self.httpd.dispatcher

    def stop(self):
        """Stop the http server."""
        if not self.started_evt.is_set():
            self.logger.warning('http server was not started yet - cannot be stopped')
            return
        self._stop_requested = True
        self.httpd.shutdown()
        self.httpd.server_close()
        for thread_info in self.httpd.threads:
            if thread_info.thread.is_alive():
                thread_info.thread.join(1)
            if thread_info.thread.is_alive():
                self.logger.warning('could not end client thread for notifications from %s', thread_info.client_address)
        del self.httpd.threads[:]
        self.started_evt.clear()



"""The module contains the implementation of the BICEPS context service."""
from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073.dispatch import DispatchKey
from sdc11073.exceptions import ApiUsageError
from sdc11073.namespaces import PrefixesEnum
from sdc11073.xml_types import msg_qnames
from sdc11073.xml_types.actions import Actions
from sdc11073.xml_types.addressing_types import HeaderInformationBlock

from .serviceclientbase import GetRequestResult, HostedServiceClient

if TYPE_CHECKING:
    from concurrent.futures import Future

    from lxml import etree

    from sdc11073.consumer.manipulator import RequestManipulatorProtocol
    from sdc11073.mdib.statecontainers import AbstractMultiStateProtocol
    from sdc11073.xml_types.pm_types import InstanceIdentifier


class ContextServiceClient(HostedServiceClient):
    """Client for ContextService."""

    port_type_name = PrefixesEnum.SDC.tag('ContextService')
    notifications = (DispatchKey(Actions.EpisodicContextReport, msg_qnames.EpisodicContextReport),
                     DispatchKey(Actions.PeriodicContextReport, msg_qnames.PeriodicContextReport))

    def mk_proposed_context_object(self, descriptor_handle: str,
                                   handle: str | None = None) -> AbstractMultiStateProtocol:
        """Create a state that can be used in set_context_state operation.

        :param descriptor_handle: the descriptor for which a state shall be created or updated
        :param handle: if None, a new object with default values is created (INSERT operation).
                       Else a copy of an existing state with this handle is returned.
        :return: a context state instance
        """
        data_model = self._sdc_definitions.data_model
        mdib = self._mdib_wref()
        if mdib is None:
            raise ApiUsageError('no mdib information')
        context_entity = mdib.entities.by_handle(descriptor_handle)
        if handle is None:
            cls = data_model.get_state_container_class(context_entity.descriptor.STATE_QNAME)
            obj = cls(descriptor_container=context_entity.descriptor)
            obj.Handle = descriptor_handle  # this indicates that this is a new context state
        else:
            _obj = context_entity.states[handle]
            obj = _obj.mk_copy()
        return obj

    def set_context_state(self, operation_handle: str,
                          proposed_context_states: list,
                          request_manipulator: RequestManipulatorProtocol | None = None) -> Future:
        """Send a SetContextState request.

        :param operation_handle: the descriptor for which a state shall be created or updated
        :param proposed_context_states: list of proposed context states
        :param request_manipulator: see documentation of RequestManipulatorProtocol
        :return: a concurrent.futures.Future object
        """
        data_model = self._sdc_definitions.data_model
        tmp = ', '.join([f'{st.__class__.__name__}(DescriptorHandle={st.DescriptorHandle}, handle={st.Handle})'
                         for st in proposed_context_states])
        self._logger.info('set_context_state {}', tmp)  # noqa: PLE1205
        request = data_model.msg_types.SetContextState()
        request.OperationHandleRef = operation_handle
        request.ProposedContextState.extend(proposed_context_states)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        return self._call_operation(message, request_manipulator=request_manipulator)

    def get_context_states(self, handles: list[str] | None = None,
                           request_manipulator: RequestManipulatorProtocol | None = None) -> GetRequestResult:
        """Send a GetContextStates request.

        :param handles: a list of handles
        :param request_manipulator: see documentation of RequestManipulatorProtocol
        :return: result of the call
        """
        data_model = self._sdc_definitions.data_model
        request = data_model.msg_types.GetContextStates()
        if handles is not None:
            request.HandleRef.extend(handles)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        received_message_data = self.post_message(message, request_manipulator=request_manipulator)
        cls = received_message_data.msg_reader.msg_types.GetContextStatesResponse
        report = cls.from_node(received_message_data.p_msg.msg_node)
        return GetRequestResult(received_message_data, report)

    def get_context_state_by_identification(self, identifications: list[InstanceIdentifier],
                                            context_type: etree.QName | None = None,
                                            request_manipulator: RequestManipulatorProtocol | None = None) \
            -> GetRequestResult:
        """Send a GetContextStatesByIdentification request.

        :param identifications: list of identifiers
        :param context_type: Type to query
        :param request_manipulator: see documentation of RequestManipulatorProtocol
        :return:
        """
        data_model = self._sdc_definitions.data_model
        request = data_model.msg_types.GetContextStatesByIdentification()
        if identifications is not None:
            request.Identification.extend(identifications)
        request.ContextType = context_type
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        received_message_data = self.post_message(message, request_manipulator=request_manipulator)
        cls = data_model.msg_types.GetContextStatesByIdentificationResponse
        report = cls.from_node(received_message_data.p_msg.msg_node)
        return GetRequestResult(received_message_data, report)

    def get_context_state_by_filter(self, filters: list[str],
                                    request_manipulator: RequestManipulatorProtocol | None = None) -> GetRequestResult:
        """Send a GetContextStatesByFilter request.

        :param filters: List of XPath expressions.
        :param request_manipulator: see documentation of RequestManipulatorProtocol
        :return: GetRequestResult
        """
        data_model = self._sdc_definitions.data_model
        request = data_model.msg_types.GetContextStatesByFilter()
        request.Filter.extend(filters)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        received_message_data = self.post_message(message, request_manipulator=request_manipulator)
        cls = data_model.msg_types.GetContextStatesByFilterResponse
        report = cls.from_node(received_message_data.p_msg.msg_node)
        return GetRequestResult(received_message_data, report)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable
from threading import Lock
from sdc11073 import loghelper

if TYPE_CHECKING:
    from .soapclient import SoapClientProtocol

    _SoapClientFactory = Callable[[str, list[str]], SoapClientProtocol]


class _SoapClientEntry:
    def __init__(self, soap_client: SoapClientProtocol | None, usr_ident: Any):
        self.soap_client = soap_client
        self.usr_idents = [usr_ident]


class SoapClientPool:
    """Pool of soap clients with reference count."""

    def __init__(self, soap_client_factory: _SoapClientFactory, log_prefix: str):
        self._soap_client_factory = soap_client_factory
        self._soap_clients: dict[str, _SoapClientEntry] = {}
        self._logger = loghelper.get_logger_adapter('sdc.device.soap_client_pool', log_prefix)
        self.async_loop_subscr_mgr = None  # is set by async subscription manager
        self._lock = Lock()

    def get_soap_client(self, netloc: str,
                        accepted_encodings: list[str],
                        usr_ident: Any) -> SoapClientProtocol:
        """Return a soap client for netloc.

        Method creates a new soap client if it did not exist yet.
        It also associates the user_ref (subscription) to the network location.
        """
        self._logger.debug('requested soap client for netloc {}', netloc)  # noqa: PLE1205
        with self._lock:
            entry = self._soap_clients.get(netloc)
            if entry is None:
                soap_client = self._soap_client_factory(netloc, accepted_encodings)
                entry = _SoapClientEntry(soap_client, usr_ident)
                self._soap_clients[netloc] = entry
            elif usr_ident not in entry.usr_idents:
                entry.usr_idents.append(usr_ident)
            return entry.soap_client

    def forget_usr(self, netloc: str, usr_ident: Any) -> None:
        """Remove the user reference from the network location.

        If no more associations exist, the soap connection gets closed and the soap client deleted.
        """
        self._logger.info('forget_usr called for netloc {} usr_ident {}', netloc, usr_ident)  # noqa: PLE1205
        with self._lock:
            entry = self._soap_clients.get(netloc)
            if entry is None or len(entry.usr_idents) == 0:
                # nothing to do
                return
            if usr_ident in entry.usr_idents:
                entry.usr_idents.remove(usr_ident)
                self._logger.info('forget user ref for netloc {}, {} user refs remaining',  # noqa: PLE1205
                                  netloc, len(entry.usr_idents))
            if len(entry.usr_idents) == 0:
                if entry.soap_client is not None:
                    self._logger.info('close soap client for netloc {}', netloc)  # noqa: PLE1205
                    if self.async_loop_subscr_mgr is None:
                        entry.soap_client.close()
                    else:
                        self.async_loop_subscr_mgr.run_coro(entry.soap_client.async_close())
                self._soap_clients.pop(netloc)

    def close_all(self):
        """Close all connections."""
        with self._lock:
            for entry in self._soap_clients.values():
                entry.soap_client.close()
            self._soap_clients = {}
        if self.async_loop_subscr_mgr is not None:
            self.async_loop_subscr_mgr.stop()



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING
from urllib.parse import urlparse

from sdc11073 import loghelper
from sdc11073.exceptions import ApiUsageError

if TYPE_CHECKING:
    from concurrent.futures import Future

    from lxml import etree

    from sdc11073.consumer.consumerimpl import SdcConsumer
    from sdc11073.consumer.manipulator import RequestManipulatorProtocol
    from sdc11073.consumer.operations import OperationsManagerProtocol
    from sdc11073.dispatch import DispatchKey
    from sdc11073.mdib.consumermdib import ConsumerMdib
    from sdc11073.namespaces import PrefixNamespace
    from sdc11073.pysoap.msgfactory import CreatedMessage
    from sdc11073.pysoap.msgreader import MdibVersionGroupReader, MessageReader, ReceivedMessage
    from sdc11073.pysoap.soapclient import SoapClientProtocol
    from sdc11073.pysoap.soapenvelope import ReceivedSoapMessage
    from sdc11073.xml_types.addressing_types import EndpointReferenceType
    from sdc11073.xml_types.basetypes import MessageType
    from sdc11073.xml_types.mex_types import HostedServiceType


class GetRequestResult:
    """Like ReceivedMessage, but plus result (StateContainers, DescriptorContainers, ...)."""

    def __init__(self, received_message: ReceivedMessage, result: MessageType):
        self._received_message = received_message
        self._result = result

    @property
    def msg_reader(self) -> MessageReader:
        """Return the responsible message reader."""
        return self._received_message.msg_reader

    @property
    def p_msg(self) -> ReceivedSoapMessage:
        """Return the received ReceivedSoapMessage."""
        return self._received_message.p_msg

    @property
    def mdib_version_group(self) -> MdibVersionGroupReader:
        """Return the mdib version group of the result."""
        return self._received_message.mdib_version_group

    @property
    def action(self) -> str | None:
        """Return the action of the result."""
        return self._received_message.action

    @property
    def msg_name(self) -> etree.QName:
        """Return the QName of message body."""
        return self._received_message.q_name.localname

    @property
    def result(self) -> MessageType:
        """Return the converted message body."""
        return self._result


class HostedServiceClient:
    """Base class of clients that call hosted services of a dpws device."""

    additional_namespaces: tuple[PrefixNamespace] = ()  # for special namespaces
    # notifications is a list of notifications that a HostedServiceClient handles (for dispatching of subscribed data).
    # Derived classes will set this class variable accordingly:
    notifications: tuple[DispatchKey] = tuple()

    def __init__(self, sdc_consumer: SdcConsumer,
                 soap_client: SoapClientProtocol,
                 dpws_hosted: HostedServiceType,
                 port_type: etree.QName):
        """Construct a HostedServiceClient.

        :param sdc_consumer:
        :param soap_client:
        :param dpws_hosted:
        :param port_type:
        """
        self.soap_client = soap_client
        self._sdc_client = sdc_consumer
        self._sdc_definitions = sdc_consumer.sdc_definitions
        self._msg_factory = sdc_consumer.msg_factory
        self.log_prefix = sdc_consumer.log_prefix
        self.dpws_hosted: HostedServiceType = dpws_hosted
        self.endpoint_reference: EndpointReferenceType = dpws_hosted.EndpointReference[0]
        self._url = urlparse(self.endpoint_reference.Address)
        self._porttype = port_type
        self._logger = loghelper.get_logger_adapter(f'sdc.client.{port_type}', self.log_prefix)
        self._operations_manager = None
        self._mdib_wref = None
        ns_helper = self._sdc_definitions.data_model.ns_helper
        self._nsmapper = ns_helper

    def register_mdib(self, mdib: ConsumerMdib | None):
        """Client sometimes must know the mdib data (e.g. Set service, activate method)."""
        if mdib is not None and self._mdib_wref is not None:
            raise ApiUsageError(f'Client "{self._porttype}" has already an registered mdib')
        self._mdib_wref = None if mdib is None else weakref.ref(mdib)

    def set_operations_manager(self, operations_manager: OperationsManagerProtocol):
        """Set the operations manager."""
        self._operations_manager = operations_manager

    def _call_operation(self, message: ReceivedMessage,
                        request_manipulator: RequestManipulatorProtocol | None = None) -> Future:
        return self._operations_manager.call_operation(self, message, request_manipulator)

    def get_available_subscriptions(self) -> tuple[DispatchKey]:
        """Return the notifications that a service offers.

        Each returned DispatchKey contains the action for the subscription and the message type that corresponds to it.
        """
        return self.notifications

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} "{self._porttype}" endpoint = {self.endpoint_reference}'

    def post_message(self, created_message: CreatedMessage,
                     msg: str | None = None,
                     request_manipulator: RequestManipulatorProtocol | None = None,
                     validate: bool = True) -> ReceivedMessage:
        """Post the created message to provider."""
        msg = msg or created_message.p_msg.payload_element.tag.split('}')[-1]

        response = self.soap_client.post_message_to(self._url.path, created_message, msg=msg,
                                                    request_manipulator=request_manipulator,
                                                    validate=validate)
        if response is None:
            raise ValueError('expect a response, got None')
        return response



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar

if TYPE_CHECKING:
    from types import ModuleType

    from lxml import etree

    from .mdib.descriptorcontainers import AbstractDescriptorProtocol
    from .mdib.statecontainers import AbstractStateProtocol
    from .namespaces import NamespaceHelper


class ProtocolsRegistry(type):
    """base class that has the only purpose to register classes that use this as metaclass."""

    protocols: ClassVar[list[type[BaseDefinitions]]] = []

    def __new__(cls, name: str, *arg, **kwarg):
        new_cls: ProtocolsRegistry = super().__new__(cls, name, *arg, **kwarg)
        if name != 'BaseDefinitions':  # ignore the base class itself
            new_cls: type[BaseDefinitions]
            cls.protocols.append(new_cls)
        return new_cls


class AbstractDataModel(ABC):
    """Abstract base class for DataModelProtocol implementation."""

    @abstractmethod
    def get_descriptor_container_class(self, type_qname: etree.QName) -> type[AbstractDescriptorProtocol]:
        """Get the class that represents a BICEPS descriptor entity with given QName."""

    def mk_descriptor_container(self, type_qname: etree.QName, handle: str, parent_descriptor: Any) -> Any:
        """Create an instance that represents a BICEPS entity with given QName."""
        cls = self.get_descriptor_container_class(type_qname)
        if parent_descriptor is not None:
            ret = cls(handle, parent_descriptor.Handle)
            ret.set_source_mds(parent_descriptor.source_mds)
        else:
            ret = cls(handle, None)
        return ret

    @abstractmethod
    def get_state_container_class(self, type_qname: etree.QName) -> type[AbstractStateProtocol]:
        """Get the class that represents a BICEPS state entity with given QName."""

    def get_state_class_for_descriptor(
            self, descriptor_container: AbstractDescriptorProtocol) -> type[AbstractStateProtocol]:
        """Get the corresponding state class for a descriptor."""
        state_class_qtype = descriptor_container.STATE_QNAME
        if state_class_qtype is None:
            raise TypeError(f'No state association for {descriptor_container.__class__.__name__}')
        return self.get_state_container_class(state_class_qtype)

    def mk_state_container(self, descriptor_container: AbstractDescriptorProtocol) -> AbstractStateProtocol:
        """Create an instance that represents a BICEPS entity with given QName."""
        cls = self.get_state_class_for_descriptor(descriptor_container)
        if cls is None:
            raise TypeError(
                f'No state container class for descr={descriptor_container.__class__.__name__}, '
                f'name={descriptor_container.NODETYPE}')
        return cls(descriptor_container)

    @property
    @abstractmethod
    def pm_types(self) -> ModuleType:
        """Get a module with participant model types."""

    @property
    @abstractmethod
    def pm_names(self) -> ModuleType:
        """Get a module with all qualified names of the BICEPS participant model."""

    @property
    @abstractmethod
    def msg_types(self) -> ModuleType:
        """Get a module with message model types."""

    @property
    @abstractmethod
    def msg_names(self) -> ModuleType:
        """Get a module with all qualified names of the BICEPS message model."""

    @property
    @abstractmethod
    def ns_helper(self) -> NamespaceHelper:
        """Gives access to a NamespaceHelper."""


class BaseDefinitions(metaclass=ProtocolsRegistry):
    """Base class for central definitions used by SDC.

    It defines namespaces and handlers for the protocol.
    Derive from this class in order to define different protocol handling.
    """

    # set the following values in derived classes:
    MedicalDeviceType: etree.QName = None  # a QName, needed for types_match method
    ActionsNamespace: str = None  # needed for wsdl generation
    PortTypeNamespace: str = None  # needed for wsdl generation
    MedicalDeviceTypesFilter: tuple[etree.QName] | None = None  # QNames that are used / expected in "types" of wsdiscovery
    Actions = None
    data_model: AbstractDataModel = None

    @classmethod
    def types_match(cls, types: list[etree.QName]) -> bool:
        """Check if this definition can be used for the provided types."""
        return cls.MedicalDeviceType in types



"""The module contains protocol definitions for the entity interface of mdib."""
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Protocol, Union

if TYPE_CHECKING:
    from collections.abc import Iterable

    from lxml.etree import QName

    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorProtocol
    from sdc11073.mdib.statecontainers import AbstractMultiStateProtocol, AbstractStateProtocol
    from sdc11073.xml_types.pm_types import CodedValue, Coding


class EntityProtocol(Protocol): # pragma: no cover
    """The protocol defines the interface of single-state entities."""

    descriptor: AbstractDescriptorProtocol
    state: AbstractStateProtocol
    is_multi_state: ClassVar[bool]

    @property
    def handle(self) -> str:
        """Return the handle of the descriptor."""

    @property
    def parent_handle(self) -> str | None:
        """Return the parent handle of the descriptor."""

    @property
    def node_type(self) -> QName:
        """Return the node type of the descriptor."""

    def update(self):
        """Update entity with current mdib data."""


class MultiStateEntityProtocol(Protocol): # pragma: no cover
    """The protocol defines the interface of multi-state entities."""

    descriptor: AbstractDescriptorProtocol
    states: dict[str, AbstractMultiStateProtocol]  # key is the Handle member of state
    is_multi_state: bool
    node_type: QName
    handle: str
    parent_handle: str

    def update(self):
        """Update entity with current data in mdib."""

    def new_state(self, state_handle: str | None = None) -> AbstractMultiStateProtocol:
        """Create a new state."""

EntityTypeProtocol = Union[EntityProtocol, MultiStateEntityProtocol]


class EntityGetterProtocol(Protocol): # pragma: no cover
    """The protocol defines a way to access mdib data as entities.

    This representation is independent of the internal mdib organization.
    The entities returned by the provided getter methods contain copies of the internal mdib data.
    Changing the data does not change data in the mdib.
    Use the EntityTransactionProtocol to write data back to the mdib.
    """

    def by_handle(self, handle: str) -> EntityTypeProtocol | None:
        """Return entity with given descriptor handle."""
        ...

    def by_context_handle(self, handle: str) -> MultiStateEntityProtocol | None:
        """Return multi state entity that contains a state with given handle."""
        ...

    def by_node_type(self, node_type: QName) -> list[EntityTypeProtocol]:
        """Return all entities with given node type."""
        ...

    def by_parent_handle(self, parent_handle: str | None) -> list[EntityTypeProtocol]:
        """Return all entities with given parent handle."""
        ...

    def by_coding(self, coding: Coding) -> list[EntityTypeProtocol]:
        """Return all entities with equivalent Coding."""
        ...

    def by_coded_value(self, coded_value: CodedValue) -> list[EntityTypeProtocol]:
        """Return all entities with equivalent CodedValue."""
        ...

    def items(self) -> Iterable[tuple[str, EntityTypeProtocol]]:
        """Return items of a dictionary."""
        ...

    def __len__(self) -> int:
        """Return number of entities."""
        ...


class ProviderEntityGetterProtocol(EntityGetterProtocol): # pragma: no cover
    """The protocol adds the new_entity method to EntityGetterProtocol."""

    def new_entity(self,
                   node_type: QName,
                   handle: str,
                   parent_handle: str | None) -> EntityTypeProtocol:
        """Create an entity."""



"""Implementation of component provider functionality."""
from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073.provider.operations import ExecuteResult

from . import providerbase

if TYPE_CHECKING:
    from sdc11073.mdib.descriptorcontainers import AbstractOperationDescriptorProtocol
    from sdc11073.provider.operations import ExecuteParameters, OperationDefinitionBase

    from .providerbase import OperationClassGetter


class GenericSetComponentStateOperationProvider(providerbase.ProviderRole):
    """Class is responsible for SetComponentState operations."""

    def make_operation_instance(self,
                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:
        """Return an operation definition instance for this operation or None.

        Can handle following case:
        operation_descriptor_container is a SetComponentStateOperationDescriptor and
        target is any AbstractComponentDescriptor.
        """
        pm_names = self._mdib.data_model.pm_names
        operation_target_handle = operation_descriptor_container.OperationTarget
        op_target_entity = self._mdib.entities.by_handle(operation_target_handle)

        if operation_descriptor_container.NODETYPE == pm_names.SetComponentStateOperationDescriptor:  # noqa: SIM300
            if op_target_entity.node_type in (pm_names.MdsDescriptor,
                                              pm_names.ChannelDescriptor,
                                              pm_names.VmdDescriptor,
                                              pm_names.ClockDescriptor,
                                              pm_names.ScoDescriptor,
                                              ):
                op_cls = operation_cls_getter(pm_names.SetComponentStateOperationDescriptor)
                return op_cls(operation_descriptor_container.Handle,
                              operation_target_handle,
                              self._set_component_state,
                              coded_value=operation_descriptor_container.Type)
        elif operation_descriptor_container.NODETYPE == pm_names.ActivateOperationDescriptor:  # noqa: SIM300
            #  on what can activate be called?
            if op_target_entity.node_type in (pm_names.MdsDescriptor,
                                              pm_names.ChannelDescriptor,
                                              pm_names.VmdDescriptor,
                                              pm_names.ScoDescriptor,
                                              ):
                # no generic handler to be called!
                op_cls = operation_cls_getter(pm_names.ActivateOperationDescriptor)
                return op_cls(operation_descriptor_container.Handle,
                              operation_target_handle,
                              self._do_nothing,
                              coded_value=operation_descriptor_container.Type)
        return None

    def _set_component_state(self, params: ExecuteParameters) -> ExecuteResult:
        """Handle SetComponentState operation (ExecuteHandler)."""
        value = params.operation_request.argument
        params.operation_instance.current_value = value
        with self._mdib.component_state_transaction() as mgr:
            for proposed_state in value:
                entity = self._mdib.entities.by_handle(proposed_state.DescriptorHandle)
                if entity.state.is_component_state:
                    self._logger.info('updating %s with proposed component state', entity.state)
                    entity.state.update_from_other_container(
                        proposed_state, skipped_properties=['StateVersion', 'DescriptorVersion'])
                    mgr.write_entity(entity)
                else:
                    self._logger.warning(
                        '_set_component_state operation: ignore invalid referenced type %s in operation',
                        entity.node_type.localname)
        return ExecuteResult(params.operation_instance.operation_target_handle,
                             self._mdib.data_model.msg_types.InvocationState.FINISHED)

    def _do_nothing(self, params: ExecuteParameters) -> ExecuteResult:
        return ExecuteResult(params.operation_instance.operation_target_handle,
                             self._mdib.data_model.msg_types.InvocationState.FINISHED)



"""implementation of ws discovery."""

from __future__ import annotations

import logging
import random
import time
from enum import Enum
from typing import TYPE_CHECKING, Callable
from urllib.parse import unquote, urlsplit

from sdc11073 import network
from sdc11073.definitions_sdc import SdcV1Definitions
from sdc11073.exceptions import ApiUsageError
from sdc11073.namespaces import default_ns_helper as nsh
from sdc11073.wsdiscovery import networkingthread
from sdc11073.xml_types import wsd_types
from sdc11073.xml_types.addressing_types import HeaderInformationBlock

from .common import MULTICAST_IPV4_ADDRESS, MULTICAST_PORT, message_factory
from .service import Service

if TYPE_CHECKING:
    import ipaddress
    from collections.abc import Iterable, Sequence
    from logging import Logger

    from lxml import etree

    from sdc11073.location import SdcLocation
    from sdc11073.pysoap.msgfactory import CreatedMessage
    from sdc11073.pysoap.msgreader import ReceivedMessage
    from sdc11073.xml_types.msg_types import MessageType

WSA_ANONYMOUS = nsh.WSA.namespace + '/anonymous'
ADDRESS_ALL = 'urn:docs-oasis-open-org:ws-dd:ns:discovery:2009:01'  # format acc to RFC 2141

NS_D = nsh.WSD.namespace

# AppSequence node is mandatory acc. to text is wsdicovery spec, but in reality this is sometimes missing.
# An application can decide to modify this value if it seems appropriate.
# If allow_missing_app_sequence is True, the message is accepted and an InstanceId of 0 is used.
# If allow_missing_app_sequence is False and the element is missing, the whole message will be ignored.
allow_missing_app_sequence = False


class MatchBy(str, Enum):
    """Different Match Options."""

    ldap = NS_D + '/ldap'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/ldap"
    uri = NS_D + '/rfc3986'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/rfc3986"
    uuid = NS_D + '/uuid'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/uuid"
    strcmp = NS_D + '/strcmp0'  # "http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01/strcmp0"


def types_info(types: Iterable[etree.QName] | None) -> list[str] | None:
    """Make printable strings from list of QNames (helper method for logging)."""
    return [str(t) for t in types] if types else types


def match_scope(my_scope: str, other_scope: str, match_by: MatchBy | str | None) -> bool:
    """match_scope checks if my_scope matches other_scope by applying the algorithm defined by match_by.

    match_scope correctly handles "%2F" (== '/') encoded values.
    """
    if match_by in (MatchBy.ldap, MatchBy.uri, MatchBy.uuid, '', None):
        my_scope = urlsplit(my_scope)
        other_scope = urlsplit(other_scope)
        if (
            my_scope.scheme.lower() != other_scope.scheme.lower()
            or my_scope.netloc.lower() != other_scope.netloc.lower()
        ):
            return False
        if my_scope.path == other_scope.path:
            return True
        src_path_elements = my_scope.path.split('/')
        target_path_elements = other_scope.path.split('/')
        src_path_elements = [unquote(elem) for elem in src_path_elements]
        target_path_elements = [unquote(elem) for elem in target_path_elements]
        if len(src_path_elements) > len(target_path_elements):
            return False
        return all(target_path_elements[i] == elem for i, elem in enumerate(src_path_elements))
    if match_by == MatchBy.strcmp:
        return my_scope == other_scope
    return False


def match_type(type1: etree.QName, type2: etree.QName) -> bool:
    """Check if namespace and localname are identical."""
    return type1.namespace == type2.namespace and type1.localname == type2.localname


def _is_type_in_list(ttype: etree.QName, types: list[etree.QName]) -> bool:
    return any(match_type(ttype, entry) for entry in types)


def _is_scope_in_list(uri: str, match_by: str, srv_sc: wsd_types.ScopesType) -> bool:
    # returns True if every entry in scope.text is also found in srv_sc.text
    # all entries are URIs
    if srv_sc is None:
        return False
    return any(match_scope(uri, entry, match_by) for entry in srv_sc.text)


def matches_filter(service: Service, types: Iterable[etree.QName] | None, scopes: wsd_types.ScopesType | None) -> bool:
    """Check if service matches the types and scopes."""
    if types is not None:
        for ttype in types:
            if not _is_type_in_list(ttype, service.types):
                return False
    if scopes is not None:
        for uri in scopes.text:
            if not _is_scope_in_list(uri, scopes.MatchBy, service.scopes):
                return False
    return True


def filter_services(
    services: Iterable[Service],
    types: Iterable[etree.QName] | None,
    scopes: wsd_types.ScopesType | None,
) -> Sequence[Service]:
    """Filter services that match types and scopes."""
    return [service for service in services if matches_filter(service, types, scopes)]


def _mk_wsd_soap_message(header_info: HeaderInformationBlock, payload: MessageType) -> CreatedMessage:
    # use discovery specific namespaces
    return message_factory.mk_soap_message(
        header_info,
        payload,
        ns_list=[nsh.S12, nsh.WSA, nsh.WSD],
        use_defaults=False,
    )


class WSDiscovery:
    """UDP based discovery."""

    # these flags control which data is included in ProbeResponse messages.
    PROBEMATCH_EPR = True
    PROBEMATCH_TYPES = True
    PROBEMATCH_SCOPES = True
    PROBEMATCH_XADDRS = True

    def __init__(
        self,
        ip_address: str | ipaddress.IPv4Address,
        logger: Logger | None = None,
        multicast_port: int = MULTICAST_PORT,
        multicast_ttl: int = 128,
    ):
        """Create a WsDiscovery instance.

        :param ip_address: network adapter to bind to
        :param logger: use this logger. if None a logger 'sdc.discover' is created.
        :param multicast_port: defaults to MULTICAST_PORT.
               If port is changed, instance will not be able to communicate with implementations
               that use the correct port (which is the default MULTICAST_PORT)!
        :param multicast_ttl: set the socket option IP_MULTICAST_TTL to this value
        """
        self._adapter = network.get_adapter_containing_ip(ip_address)
        self._networking_thread = None
        self._addrs_monitor_thread = None
        self._server_started = False
        self._remote_services = {}
        self._local_services = {}
        self._remote_service_hello_callback = None
        self._remote_service_hello_callback_types_filter = None
        self._remote_service_hello_callback_scopes_filter = None
        self._remote_service_bye_callback = None
        self._remote_service_resolve_match_callback: Callable[[Service], None] | None = None  # B.D.
        self._on_probe_callback = None
        self._probe_match_callback: Callable[[Sequence[Service]], None] | None = None

        self._logger = logger or logging.getLogger('sdc.discover')
        self.multicast_port = multicast_port
        self.mutlticast_ttl = multicast_ttl

    def start(self):
        """Start the discovery server - should be called before using other functions."""
        if not self._server_started:
            self._start_threads()
            self._server_started = True

    def stop(self):
        """Clean up and stop the discovery server."""
        if self._server_started:
            self.clear_remote_services()
            self.clear_local_services()

            self._stop_threads()
            self._server_started = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args, **kwargs):  # noqa: ANN002, ANN003
        self.stop()

    def search_services(
        self,
        types: Iterable[etree.QName] | None = None,
        scopes: wsd_types.ScopesType | None = None,
        timeout: int | float | None = 5,  # noqa: PYI041
        repeat_probe_interval: int | None = 3,
    ) -> Sequence[Service]:
        """Search for services that match given types and scopes.

        :param types: list of types that a service must have (all of them), no filtering if value is None
        :param scopes:scopes to search for, no filtering if value is None
        :param timeout: total duration of search
        :param repeat_probe_interval: send another probe message after x seconds
        :return:
        """
        if not self._server_started:
            raise RuntimeError('Server not started')

        types = list(types) if types is not None else None
        start = time.monotonic()
        end = start + timeout
        now = time.monotonic()
        while now < end:
            self._send_probe(types, scopes)
            if now + repeat_probe_interval <= end:
                time.sleep(repeat_probe_interval)
            elif now < end:
                time.sleep(end - now)
            now = time.monotonic()
        return filter_services(list(self._remote_services.values()), types, scopes)

    def get_found_remote_services(
        self, types: Iterable[etree.QName] | None = None, scopes: wsd_types.ScopesType | None = None,
    ) -> Sequence[Service]:
        """Get currently known remote services that match given types and scopes."""
        return filter_services(list(self._remote_services.values()), types, scopes)

    def search_sdc_services(
        self,
        scopes: wsd_types.ScopesType | None = None,
        timeout: int | float | None = 5,  # noqa: PYI041
        repeat_probe_interval: int | None = 3,
    ) -> Sequence[Service]:
        """Search for sdc services that match given scopes.

        :param scopes: scopes to search for, no scopes filtering if value is None
        :param timeout: total duration of search
        :param repeat_probe_interval: send another probe message after x seconds
        :return:
        """
        return self.search_services(SdcV1Definitions.MedicalDeviceTypesFilter, scopes, timeout, repeat_probe_interval)

    def search_multiple_types(
        self,
        types_list: list[list[etree.QName]],
        scopes: wsd_types.ScopesType | None = None,
        timeout: int | float | None = 10,  # noqa: PYI041
        repeat_probe_interval: int | None = 3,
    ) -> list[Service]:
        """Search for services given the list of TYPES and SCOPES in a given timeout.

        It returns services that match at least one of the types (OR condition).
        Can be used to search for devices that support Biceps Draft6 and Final with one search.
        :param types_list:
        :param scopes:
        :param timeout: total duration of search
        :param repeat_probe_interval: send another probe message after x seconds.
        """
        if not self._server_started:
            raise ApiUsageError('Server not started')

        start = time.monotonic()
        end = start + timeout
        now = time.monotonic()
        while now < end:
            for _type in types_list:
                self._send_probe(_type, scopes)
            now = time.monotonic()
            if now + repeat_probe_interval <= end:
                time.sleep(repeat_probe_interval)
            elif now < end:
                time.sleep(end - now)
        # prevent possible duplicates by adding them to a dictionary by epr
        result = {}
        for _type in types_list:
            tmp = filter_services(list(self._remote_services.values()), _type, scopes)
            for srv in tmp:
                result[srv.epr] = srv
        return list(result.values())

    def search_sdc_device_services_in_location(self, sdc_location: SdcLocation, timeout: int = 3) -> list[Service]:
        """Search for all sdc devices (no scopes filter applied), then filter locally for location."""
        services = self.search_sdc_services(timeout=timeout)
        return sdc_location.filter_services_inside(services)

    def publish_service(self, epr: str, types: list[etree.QName], scopes: wsd_types.ScopesType, x_addrs: list[str]):
        """Publish a service with the given TYPES, SCOPES and XAddrs (service addresses).

        if x_addrs contains item, which includes {ip} pattern, one item per IP address will be sent
        """
        if not self._server_started:
            raise ApiUsageError('Server not started')

        metadata_version = self._local_services[epr].metadata_version + 1 if epr in self._local_services else 1
        instance_id = str(random.randint(1, 0xFFFFFFFF))
        service = Service(types, scopes, x_addrs, epr, instance_id, metadata_version=metadata_version)
        self._logger.info('publishing %r', service)
        self._local_services[epr] = service
        self._send_hello(service)

    def clear_remote_services(self):
        """Clear remotely discovered services."""
        self._remote_services.clear()

    def clear_local_services(self):
        """Send Bye messages for the services and remove them."""
        for service in self._local_services.values():
            self._send_bye(service)
        self._local_services.clear()

    def clear_service(self, epr: str):
        """Clear local service with given epr."""
        service = self._local_services[epr]
        self._send_bye(service)
        del self._local_services[epr]

    def get_active_addresses(self) -> list[str]:
        """Get active addresses."""
        return [str(self._adapter.ip)]

    def set_remote_service_hello_callback(
        self,
        callback: Callable[[str, Service], None] | None,
        types: list[etree.QName] | None = None,
        scopes: wsd_types.ScopesType | None = None,
    ):
        """Set callback, which will be called when new service appeared online and sent Hello message.

        typesFilter and scopesFilter might be list of types and scopes.
        If filter is set, callback is called only for Hello messages,
        which match filter

        Set None to disable callback
        """
        self._remote_service_hello_callback = callback
        self._remote_service_hello_callback_types_filter = types
        self._remote_service_hello_callback_scopes_filter = scopes

    def set_remote_service_bye_callback(self, callback: Callable[[str, str], None] | None):
        """Set callback, which will be called when a bye message is received.

        Set to None to disable callback.
        """
        self._remote_service_bye_callback = callback

    def set_remote_service_resolve_match_callback(self, callback: Callable[[Service], None] | None):
        """Set callback, which will be called when a resolve match message is received.

        Set to None to disable callback.
        """
        self._remote_service_resolve_match_callback = callback

    def set_on_probe_callback(self, callback: Callable[[str, wsd_types.ProbeType], None] | None):
        """Set callback, which will be called when a probe message is received.

        Set to None to disable callback.
        """
        self._on_probe_callback = callback

    def set_on_probe_matches_callback(self, callback: Callable[[Sequence[Service]], None] | None):
        """Set callback, which will be called when a probe matches message is received."""
        self._probe_match_callback = callback

    def _add_remote_service(self, service: Service):
        if not service.epr:
            self._logger.info('service without epr, ignoring it! %r', service)
            return
        already_known_service = self._remote_services.get(service.epr)
        if not already_known_service:
            self._remote_services[service.epr] = service
            self._logger.info('new remote epr="%s" x_addrs=%r', service.epr, service.x_addrs)
            return

        if service.metadata_version == already_known_service.metadata_version:
            self._logger.debug(
                'update remote service: remote Service %s; MetadataVersion: %d',
                service.epr,
                service.metadata_version,
            )
            if len(service.x_addrs) > len(already_known_service.x_addrs):
                already_known_service.x_addrs = service.x_addrs
            if service.scopes is not None:
                already_known_service.scopes = service.scopes
            if service.types is not None:
                already_known_service.types = service.types
        elif service.metadata_version > already_known_service.metadata_version:
            self._logger.info(
                'remote Service %s:\n    updated MetadataVersion\n      updated: %d\n      existing: %d',
                service.epr,
                service.metadata_version,
                already_known_service.metadata_version,
            )
            self._remote_services[service.epr] = service
        else:
            self._logger.debug(
                '_add_remote_service: remote Service %s:\n    outdated MetadataVersion\n      '
                'outdated: %d\n      existing: %d',
                service.epr,
                service.metadata_version,
                already_known_service.metadata_version,
            )

    def _remove_remote_service(self, epr: str):
        if epr in self._remote_services:
            del self._remote_services[epr]

    def _handle_received_hello(self, received_message: ReceivedMessage, addr_from: str):
        app_sequence_node = received_message.p_msg.header_node.find(nsh.WSD.tag('AppSequence'))
        if app_sequence_node is None:
            if allow_missing_app_sequence:
                instance_id = 0
            else:
                self._logger.debug('received Hello from %r without AppSequence, ignoring it', addr_from)
                return
        else:
            app_sequence = wsd_types.AppSequenceType.from_node(app_sequence_node)
            instance_id = app_sequence.InstanceId
        hello = wsd_types.HelloType.from_node(received_message.p_msg.msg_node)
        epr = hello.EndpointReference.Address
        scopes = hello.Scopes
        service = Service(hello.Types, scopes, hello.XAddrs, epr, instance_id, metadata_version=hello.MetadataVersion)
        self._add_remote_service(service)
        if not hello.XAddrs:  # B.D.
            self._logger.debug('%s(%s) has no Xaddr, sending resolve message', epr, addr_from)
            self._send_resolve(epr)
        if self._remote_service_hello_callback is not None:
            if matches_filter(
                service,
                self._remote_service_hello_callback_types_filter,
                self._remote_service_hello_callback_scopes_filter,
            ):
                self._remote_service_hello_callback(addr_from, service)

    def _handle_received_probe(self, received_message: ReceivedMessage, addr_from: str):
        probe = wsd_types.ProbeType.from_node(received_message.p_msg.msg_node)
        scopes = probe.Scopes
        services = filter_services(self._local_services.values(), probe.Types, scopes)
        if services:
            self._send_probe_match(services, received_message.p_msg.header_info_block.MessageID, addr_from)
        if self._on_probe_callback is not None:
            self._on_probe_callback(addr_from, probe)

    def _handle_received_probe_matches(self, received_message: ReceivedMessage, addr_from: str):
        app_sequence_node = received_message.p_msg.header_node.find(nsh.WSD.tag('AppSequence'))
        if app_sequence_node is None:
            if allow_missing_app_sequence:
                instance_id = 0
            else:
                self._logger.debug('received ProbeMatches from %r without AppSequence, ignoring it', addr_from)
                return
        else:
            app_sequence = wsd_types.AppSequenceType.from_node(app_sequence_node)
            instance_id = app_sequence.InstanceId
        probe_matches = wsd_types.ProbeMatchesType.from_node(received_message.p_msg.msg_node)
        self._logger.debug('handle_received_message: len(ProbeMatch) = %d', len(probe_matches.ProbeMatch))
        services: list[Service] = []
        for match in probe_matches.ProbeMatch:
            epr = match.EndpointReference.Address
            scopes = match.Scopes
            service = Service(
                match.Types,
                scopes,
                match.XAddrs,
                epr,
                instance_id,
                metadata_version=match.MetadataVersion,
            )
            services.append(service)
            self._add_remote_service(service)
            if match.XAddrs is None or len(match.XAddrs) == 0:
                self._logger.info('%s(%s) has no Xaddr, sending resolve message', epr, addr_from)
                self._send_resolve(epr)
            elif not match.Types:
                self._logger.info('%s(%s) has no Types, sending resolve message', epr, addr_from)
                self._send_resolve(epr)
            elif not match.Scopes:
                self._logger.info('%s(%s) has no Scopes, sending resolve message', epr, addr_from)
                self._send_resolve(epr)
        if self._probe_match_callback is not None:
            self._probe_match_callback(services)

    def _handle_received_resolve(self, received_message: ReceivedMessage, addr_from: str):
        resolve = wsd_types.ResolveType.from_node(received_message.p_msg.msg_node)
        epr = resolve.EndpointReference.Address
        if epr in self._local_services:
            service = self._local_services[epr]
            self._send_resolve_match(service, received_message.p_msg.header_info_block.MessageID, addr_from)

    def _handle_received_resolve_matches(self, received_message: ReceivedMessage, addr_from: str):  # ARG002
        app_sequence_node = received_message.p_msg.header_node.find(nsh.WSD.tag('AppSequence'))
        if app_sequence_node is None:
            if allow_missing_app_sequence:
                instance_id = 0
            else:
                self._logger.debug('received ResolveMatches from %r without AppSequence, ignoring it', addr_from)
                return
        else:
            app_sequence = wsd_types.AppSequenceType.from_node(app_sequence_node)
            instance_id = app_sequence.InstanceId
        resolve_matches = wsd_types.ResolveMatchesType.from_node(received_message.p_msg.msg_node)
        match = resolve_matches.ResolveMatch
        epr = match.EndpointReference.Address
        scopes = match.Scopes
        service = Service(match.Types, scopes, match.XAddrs, epr, instance_id, metadata_version=match.MetadataVersion)
        self._add_remote_service(service)
        if self._remote_service_resolve_match_callback is not None:
            self._remote_service_resolve_match_callback(service)

    def _handle_received_bye(self, received_message: ReceivedMessage, addr_from: str):
        bye = wsd_types.ByeType.from_node(received_message.p_msg.msg_node)
        epr = bye.EndpointReference.Address
        self._remove_remote_service(epr)
        if self._remote_service_bye_callback is not None:
            self._remote_service_bye_callback(addr_from, epr)

    def handle_received_message(self, received_message: ReceivedMessage, addr_from: str):
        """Forward received message to specific handler (dispatch by action)."""
        action = received_message.action
        self._logger.debug('handle_received_message: received %s from %s', action.split('/')[-1], addr_from)
        lookup = {
            wsd_types.HelloType.action: self._handle_received_hello,
            wsd_types.ProbeType.action: self._handle_received_probe,
            wsd_types.ProbeMatchesType.action: self._handle_received_probe_matches,
            wsd_types.ResolveType.action: self._handle_received_resolve,
            wsd_types.ResolveMatchesType.action: self._handle_received_resolve_matches,
            wsd_types.ByeType.action: self._handle_received_bye,
        }
        try:
            func: Callable[[ReceivedMessage, str], None] = lookup[action]
        except KeyError:
            self._logger.error('unknown action %s', action)  # noqa: TRY400
        else:
            func(received_message, addr_from)

    def _send_resolve_match(self, service: Service, relates_to: str, addr: str):
        self._logger.info('sending resolve match to %s', addr)
        service.increment_message_number()
        payload = wsd_types.ResolveMatchesType()
        payload.ResolveMatch = wsd_types.ResolveMatchType()
        payload.ResolveMatch.EndpointReference.Address = service.epr
        payload.ResolveMatch.MetadataVersion = service.metadata_version
        payload.ResolveMatch.Types = service.types
        payload.ResolveMatch.Scopes = service.scopes
        payload.ResolveMatch.XAddrs.extend(service.x_addrs)
        inf = HeaderInformationBlock(action=payload.action, addr_to=WSA_ANONYMOUS, relates_to=relates_to)
        app_sequence = wsd_types.AppSequenceType()
        app_sequence.InstanceId = int(service.instance_id)
        app_sequence.MessageNumber = service.message_number

        created_message = _mk_wsd_soap_message(inf, payload)
        created_message.p_msg.add_header_element(
            app_sequence.as_etree_node(nsh.WSD.tag('AppSequence'), ns_map=nsh.partial_map(nsh.WSD)),
        )
        self._networking_thread.add_outbound_message(
            created_message,
            addr[0],
            addr[1],
            networkingthread.UNICAST_REPEAT_PARAMS,
        )

    def _send_probe_match(self, services: list[Service], relates_to: str, addr: str):
        self._logger.info('sending probe match to %s for %d services', addr, len(services))
        msg_number = 1
        # send one match response for every service.
        # dpws explorer can't handle telegram otherwise if too many devices reported
        for service in services:
            payload = wsd_types.ProbeMatchesType()

            # add values to ProbeResponse acc. to flags
            epr = service.epr if self.PROBEMATCH_EPR else ''
            types = service.types if self.PROBEMATCH_TYPES else []
            scopes = service.scopes if self.PROBEMATCH_SCOPES else None
            xaddrs = service.x_addrs if self.PROBEMATCH_XADDRS else []

            probe_match = wsd_types.ProbeMatchType()
            probe_match.EndpointReference.Address = epr
            probe_match.MetadataVersion = service.metadata_version
            probe_match.Types = types
            probe_match.Scopes = scopes
            probe_match.XAddrs.extend(xaddrs)
            payload.ProbeMatch.append(probe_match)
            inf = HeaderInformationBlock(action=payload.action, addr_to=WSA_ANONYMOUS, relates_to=relates_to)
            app_sequence = wsd_types.AppSequenceType()
            app_sequence.InstanceId = int(service.instance_id)
            app_sequence.MessageNumber = msg_number

            created_message = _mk_wsd_soap_message(inf, payload)
            created_message.p_msg.add_header_element(
                app_sequence.as_etree_node(nsh.WSD.tag('AppSequence'), ns_map=nsh.partial_map(nsh.WSD)),
            )
            self._networking_thread.add_outbound_message(
                created_message,
                addr[0],
                addr[1],
                networkingthread.UNICAST_REPEAT_PARAMS,
            )

    def _send_probe(self, types: Iterable[etree.QName] | None = None, scopes: wsd_types.ScopesType | None = None):
        types = list(types) if types is not None else None  # enforce iteration
        self._logger.debug('sending probe types=%r scopes=%r', types_info(types), scopes)
        payload = wsd_types.ProbeType()
        payload.Types = types
        if scopes is not None:
            payload.Scopes = scopes

        inf = HeaderInformationBlock(action=payload.action, addr_to=ADDRESS_ALL)
        created_message = _mk_wsd_soap_message(inf, payload)
        self._networking_thread.add_outbound_message(
            created_message,
            MULTICAST_IPV4_ADDRESS,
            self.multicast_port,
            networkingthread.MULTICAST_REPEAT_PARAMS,
        )

    def _send_resolve(self, epr: str):
        self._logger.debug('sending resolve on %s', epr)
        payload = wsd_types.ResolveType()
        payload.EndpointReference.Address = epr

        inf = HeaderInformationBlock(action=payload.action, addr_to=ADDRESS_ALL)
        created_message = _mk_wsd_soap_message(inf, payload)
        self._networking_thread.add_outbound_message(
            created_message,
            MULTICAST_IPV4_ADDRESS,
            self.multicast_port,
            networkingthread.MULTICAST_REPEAT_PARAMS,
        )

    def _send_hello(self, service: Service):
        self._logger.info('sending hello on %s', service)
        service.increment_message_number()
        app_sequence = wsd_types.AppSequenceType()
        app_sequence.InstanceId = int(service.instance_id)
        app_sequence.MessageNumber = service.message_number

        payload = wsd_types.HelloType()
        payload.Types = service.types
        payload.Scopes = service.scopes
        payload.XAddrs = service.x_addrs
        payload.EndpointReference.Address = service.epr

        inf = HeaderInformationBlock(action=payload.action, addr_to=ADDRESS_ALL)

        created_message = _mk_wsd_soap_message(inf, payload)
        created_message.p_msg.add_header_element(
            app_sequence.as_etree_node(nsh.WSD.tag('AppSequence'), ns_map=nsh.partial_map(nsh.WSD)),
        )
        self._networking_thread.add_outbound_message(
            created_message,
            MULTICAST_IPV4_ADDRESS,
            self.multicast_port,
            networkingthread.MULTICAST_REPEAT_PARAMS,
        )

    def _send_bye(self, service: Service):
        self._logger.debug('sending bye on %s', service)

        bye = wsd_types.ByeType()
        bye.EndpointReference.Address = service.epr

        inf = HeaderInformationBlock(action=bye.action, addr_to=ADDRESS_ALL)

        app_sequence = wsd_types.AppSequenceType()
        app_sequence.InstanceId = int(service.instance_id)
        app_sequence.MessageNumber = service.message_number

        created_message = _mk_wsd_soap_message(inf, bye)
        created_message.p_msg.add_header_element(
            app_sequence.as_etree_node(nsh.WSD.tag('AppSequence'), ns_map=nsh.partial_map(nsh.WSD)),
        )
        self._networking_thread.add_outbound_message(
            created_message,
            MULTICAST_IPV4_ADDRESS,
            self.multicast_port,
            networkingthread.MULTICAST_REPEAT_PARAMS,
        )

    def _start_threads(self):
        if self._networking_thread is not None:
            return
        self._networking_thread = networkingthread.NetworkingThread(
            str(self._adapter.ip),
            self,
            self._logger,
            self.multicast_port,
            self.mutlticast_ttl,
        )
        self._networking_thread.start()

    def _stop_threads(self):
        if self._networking_thread is None:
            return

        self._networking_thread.schedule_stop()

        self._networking_thread.join()

        self._networking_thread = None


class WSDiscoverySingleAdapter(WSDiscovery):
    """Bind to a single adapter, identified by name."""

    def __init__(self, adapter_name: str, logger: Logger | None = None, multicast_port: int = MULTICAST_PORT):
        """WSDiscoverySingleAdapter uses an adapter name to determine the ip address.

        :param adapter_name: name of the network adapter on which the discovery is to be performed
        :param logger: use this logger. If None, 'sdc.discover' is used.
        :param multicast_port: multicast port
        """
        adapters = [adapter for adapter in network.get_adapters() if adapter.name == adapter_name]
        if not adapters:
            names = [adapter.name for adapter in network.get_adapters()]
            msg = f'No adapter named "{adapter_name}", have {names}'
            raise RuntimeError(msg)
        if len(adapters) > 1:
            msg = f'Found multiple possible ip addresses on adapter "{adapter_name}"'
            raise RuntimeError(msg)
        super().__init__(adapters[0].ip, logger, multicast_port)



from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073.xml_types.addressing_types import HeaderInformationBlock
from .subscriptionmgr_base import ActionBasedSubscription, SubscriptionsManagerBase
from sdc11073 import observableproperties
from sdc11073.httpserver.compression import CompressionHandler
from sdc11073.pysoap.soapclient import HTTPReturnCodeError
from sdc11073.xml_types import eventing_types as evt_types
from sdc11073.xml_types.dpws_types import DeviceEventingFilterDialectURI

if TYPE_CHECKING:
    from sdc11073.dispatch import RequestData
    from sdc11073 import xml_utils


class BicepsSubscription(ActionBasedSubscription):
    """ This extends ActionBasedSubscription with the ability to send notifications.
    The class is used by ActionBasedSubscriptionsManager."""

    def send_notification_report(self, body_node: xml_utils.LxmlElement, action: str):
        if not self.is_valid:
            return
        inf = HeaderInformationBlock(addr_to=self.notify_to_address,
                                     action=action,
                                     addr_from=None,
                                     reference_parameters=self.notify_ref_params)
        message = self._mk_notification_message(inf, body_node)
        try:
            soap_client = self._get_soap_client()
            roundtrip_timer = observableproperties.SingleValueCollector(soap_client, 'roundtrip_time')

            soap_client.post_message_to(self.notify_to_url.path, message,
                                        msg=f'send_notification_report {action}')
            try:
                roundtrip_time = roundtrip_timer.result(0)
                self.last_roundtrip_times.append(roundtrip_time)
                self.max_roundtrip_time = max(self.max_roundtrip_time, roundtrip_time)
            except observableproperties.CollectTimeoutError:
                pass
            self.notify_errors = 0
            self._is_connection_error = False
        except HTTPReturnCodeError:
            self.notify_errors += 1
            raise
        except Exception:  # any other exception is handled as an unreachable location (disconnected)
            self.notify_errors += 1
            self._is_connection_error = True
            raise


class ActionBasedSubscriptionsManager(SubscriptionsManagerBase):
    """This is the synchronous version of the subscription manager for all BICEPS subscriptions."""
    supported_filter_dialect = DeviceEventingFilterDialectURI.ACTION
    subscription_cls = BicepsSubscription

    def _mk_subscription_instance(self, request_data: RequestData) -> ActionBasedSubscription:
        subscribe_request = evt_types.Subscribe.from_node(request_data.message_data.p_msg.msg_node)
        filter_type = subscribe_request.Filter
        if filter_type is None:
            raise ValueError(f'No filter provided for {self.__class__.__name__}')
        if filter_type.Dialect != self.supported_filter_dialect:
            raise ValueError(
                f'Invalid filter dialect, got {filter_type.Dialect}, expect {self.supported_filter_dialect}')

        accepted_encodings = CompressionHandler.parse_header(request_data.http_header.get('Accept-Encoding'))
        return self.subscription_cls(self, subscribe_request, accepted_encodings, self.base_urls,
                                     self._max_subscription_duration, self._soap_client_pool,
                                     msg_factory=self._msg_factory, log_prefix=self._logger.log_prefix)


class PathDispatchingSubscriptionsManager(ActionBasedSubscriptionsManager):
    """This implementation uses path dispatching to identify subscriptions."""

    def _mk_subscription_instance(self, request_data: RequestData):
        subscription = super()._mk_subscription_instance(request_data)
        subscription.path_suffix = subscription.identifier_uuid.hex
        return subscription


class ReferenceParamSubscriptionsManager(ActionBasedSubscriptionsManager):
    """This implementation uses reference parameters to identify subscriptions."""

    def _mk_subscription_instance(self, request_data: RequestData):
        subscription = super()._mk_subscription_instance(request_data)
        # add  a reference parameter
        subscription.set_reference_parameter()
        return subscription



from . import xml_structure
from . import xml_structure as cp
from .addressing_types import EndpointReferenceType
from .basetypes import XMLTypeBase, ElementWithText, MessageType
from .dataconverters import DurationConverter
from .dpws_types import DeviceEventingFilterDialectURI
from sdc11073.namespaces import EventingActions
from sdc11073.namespaces import default_ns_helper

wse_tag = default_ns_helper.WSE.tag
xml_tag = default_ns_helper.XML.tag
### classes that correspond to types in eventing standard


class DeliveryType(XMLTypeBase):
    Mode = cp.AnyURIAttributeProperty("Mode")
    # in ws-eventing NotifyTo is not a mandatory element, but in sdc it is always used.
    NotifyTo = cp.SubElementProperty(wse_tag('NotifyTo'),
                                     value_class=EndpointReferenceType,
                                     default_py_value=EndpointReferenceType())
    _props = ('Mode', 'NotifyTo')


class FilterType(ElementWithText):
    Dialect = cp.AnyURIAttributeProperty('Dialect')
    any = xml_structure.AnyEtreeNodeListProperty(None, is_optional=True)  # noqa: A003
    _props = ('Dialect', 'any')


class Subscribe(MessageType):
    NODETYPE = wse_tag('Subscribe')
    action = EventingActions.Subscribe
    EndTo = cp.SubElementProperty(wse_tag('EndTo'),
                                  value_class=EndpointReferenceType,
                                  is_optional=True)
    Delivery = cp.SubElementProperty(wse_tag('Delivery'),
                                     value_class=DeliveryType,
                                     default_py_value=DeliveryType())
    Expires = cp.NodeDurationProperty(wse_tag('Expires'))
    Filter = cp.SubElementProperty(wse_tag('Filter'),
                                   value_class=FilterType,
                                   is_optional=True)
    _props = ('EndTo', 'Delivery', 'Expires', 'Filter')
    additional_namespaces = (default_ns_helper.WSE,)

    def init_end_to(self):
        self.EndTo = EndpointReferenceType()

    def set_filter(self, filter_text, dialect=DeviceEventingFilterDialectURI.ACTION):
        self.Filter = FilterType()
        self.Filter.text = filter_text
        self.Filter.Dialect = dialect


class SubscribeResponse(MessageType):
    NODETYPE = wse_tag('SubscribeResponse')
    action = EventingActions.SubscribeResponse
    SubscriptionManager = cp.SubElementProperty(wse_tag('SubscriptionManager'),
                                                value_class=EndpointReferenceType,
                                                default_py_value=EndpointReferenceType())
    Expires = cp.NodeDurationProperty(wse_tag('Expires'))
    _props = ('SubscriptionManager', 'Expires')
    additional_namespaces = (default_ns_helper.WSE,)


class Unsubscribe(MessageType):
    NODETYPE = wse_tag('Unsubscribe')
    action = EventingActions.Unsubscribe
    additional_namespaces = (default_ns_helper.WSE,)


class UnsubscribeResponse(MessageType):
    NODETYPE = wse_tag('UnsubscribeResponse')
    action = EventingActions.UnsubscribeResponse
    additional_namespaces = (default_ns_helper.WSE,)

    def as_etree_node(self, *args, **kwargs):
        # Unsubscribe has empty body
        return None


class LanguageSpecificStringType(ElementWithText):
    lang = cp.StringAttributeProperty(attribute_name=xml_tag('lang'))
    _props = ('lang',)


class SubscriptionEnd(MessageType):
    NODETYPE = wse_tag('SubscriptionEnd')
    action = EventingActions.SubscriptionEnd
    SubscriptionManager = cp.SubElementProperty(wse_tag('SubscriptionManager'),
                                                value_class=EndpointReferenceType,
                                                default_py_value=EndpointReferenceType())
    Status = cp.NodeStringProperty(wse_tag('Status'))
    Reason = cp.SubElementListProperty(wse_tag('Reason'),
                                       value_class=LanguageSpecificStringType)
    _props = ('SubscriptionManager', 'Status', 'Reason')
    additional_namespaces = (default_ns_helper.WSE,)

    def add_reason(self, text, lang=None):
        tmp = LanguageSpecificStringType()
        tmp.text = text
        tmp.lang = lang
        self.Reason.append(tmp)


class Renew(MessageType):
    NODETYPE = wse_tag('Renew')
    action = EventingActions.Renew
    Expires = cp.NodeDurationProperty(wse_tag('Expires'))
    _props = ('Expires',)
    additional_namespaces = (default_ns_helper.WSE,)


class RenewResponse(MessageType):
    NODETYPE = wse_tag('RenewResponse')
    action = EventingActions.RenewResponse
    Expires = cp.NodeDurationProperty(wse_tag('Expires'))
    _props = ('Expires',)
    additional_namespaces = (default_ns_helper.WSE,)


class GetStatus(MessageType):
    NODETYPE = wse_tag('GetStatus')
    action = EventingActions.GetStatus
    additional_namespaces = (default_ns_helper.WSE,)


class GetStatusResponse(MessageType):
    NODETYPE = wse_tag('GetStatusResponse')
    action = EventingActions.GetStatusResponse
    Expires = cp.NodeDurationProperty(wse_tag('Expires'))
    _props = ('Expires',)
    additional_namespaces = (default_ns_helper.WSE,)
